from __future__ import annotations

from abc import ABCMeta, abstractmethod
from contextvars import ContextVar
from datetime import timedelta
import enum
import logging
import math
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TYPE_CHECKING,
)
from ai.backend.common.events import (
    AbstractEvent,
    EventHandler,
    DoIdleCheckEvent,
    DoTerminateSessionEvent,
    EventDispatcher,
    EventProducer,
    ExecutionStartedEvent,
    ExecutionCancelledEvent,
    ExecutionFinishedEvent,
    ExecutionTimeoutEvent,
    SessionStartedEvent,
)

if TYPE_CHECKING:
    from aiopg.sa.connection import SAConnection
    from aiopg.sa.engine import _PoolAcquireContextManager as SAPool
    from aiopg.sa.result import RowProxy
import aioredis
import sqlalchemy as sa
import trafaret as t

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import aobject, AccessKey
import ai.backend.common.validators as tx
if TYPE_CHECKING:
    from ai.backend.common.types import AgentId, SessionId

from .distributed import GlobalTimer
from .models import kernels, keypairs, keypair_resource_policies
from .models.kernel import LIVE_STATUS
from ..gateway.defs import REDIS_LIVE_DB
if TYPE_CHECKING:
    from ..gateway.config import SharedConfig

log = BraceStyleAdapter(logging.getLogger('ai.backend.manager.idle'))


class AppStreamingStatus(enum.Enum):
    NO_ACTIVE_CONNECTIONS = 0
    HAS_ACTIVE_CONNECTIONS = 1


class BaseIdleChecker(aobject, metaclass=ABCMeta):

    name: ClassVar[str] = "base"

    _dbpool: SAPool
    _shared_config: SharedConfig
    _event_dispatcher: EventDispatcher
    _event_producer: EventProducer

    def __init__(
        self,
        dbpool: SAPool,
        shared_config: SharedConfig,
        event_dispatcher: EventDispatcher,
        event_producer: EventProducer,
    ) -> None:
        self._dbpool = dbpool
        self._shared_config = shared_config
        self._event_dispatcher = event_dispatcher
        self._event_producer = event_producer

    async def __ainit__(self) -> None:
        self._redis = await aioredis.create_redis(
            str(self._shared_config.get_redis_url(db=REDIS_LIVE_DB))
        )
        raw_config = await self._shared_config.etcd.get_prefix_dict(f"config/idle/checkers/{self.name}")
        await self.populate_config(raw_config or {})
        self.timer = GlobalTimer(
            self._redis,
            "idle_check",
            self._event_producer,
            lambda: DoIdleCheckEvent(),
            10.0,
        )
        self._evh_idle_check = self._event_dispatcher.consume(
            DoIdleCheckEvent, None, self._do_idle_check,
        )
        await self.timer.join()

    async def aclose(self) -> None:
        await self.timer.leave()
        self._event_dispatcher.unconsume(self._evh_idle_check)
        self._redis.close()
        await self._redis.wait_closed()

    @abstractmethod
    async def populate_config(self, config: Mapping[str, Any]) -> None:
        raise NotImplementedError

    @abstractmethod
    async def update_app_streaming_status(
        self,
        session_id: SessionId,
        status: AppStreamingStatus,
    ) -> None:
        pass

    async def _do_idle_check(
        self,
        context: None,
        source: AgentId,
        event: DoIdleCheckEvent,
    ) -> None:
        log.debug('do_idle_check(): triggered')
        async with self._dbpool.acquire() as conn:
            query = (
                sa.select([kernels])
                .select_from(kernels)
                .where(
                    (kernels.c.status.in_(LIVE_STATUS))
                )
            )
            result = await conn.execute(query)
            rows = await result.fetchall()
            for row in rows:
                if not (await self.check_session(row, conn)):
                    log.info(f"The {self.name} idle checker triggered termination of s:{row['id']}")
                    await self._event_producer.produce_event(
                        DoTerminateSessionEvent(row['id'], "idle-timeout")
                    )

    @abstractmethod
    async def check_session(self, session: RowProxy, dbconn: SAConnection) -> bool:
        """
        Return True if the session should be kept alive or
        return False if the session should be terminated.
        """
        return True


class TimeoutIdleChecker(BaseIdleChecker):
    """
    Checks the idleness of a session by the elapsed time since last used.
    The usage means processing of any computation requests, such as
    query/batch-mode code execution and having active service-port connections.
    """

    name: ClassVar[str] = "timeout"

    _config_iv = t.Dict({
        t.Key('threshold', default="10m"): tx.TimeDuration(),
    }).allow_extra('*')

    idle_timeout: timedelta
    _policy_cache: ContextVar[Dict[AccessKey, Optional[Mapping[str, Any]]]]
    _evhandlers: List[EventHandler[None, AbstractEvent]]

    async def __ainit__(self) -> None:
        await super().__ainit__()
        self._policy_cache = ContextVar('_policy_cache')
        d = self._event_dispatcher
        self._evhandlers = [
            d.consume(SessionStartedEvent, None, self._session_started_cb),       # type: ignore
            d.consume(ExecutionStartedEvent, None, self._execution_started_cb),   # type: ignore
            d.consume(ExecutionFinishedEvent, None, self._execution_exited_cb),   # type: ignore
            d.consume(ExecutionTimeoutEvent, None, self._execution_exited_cb),    # type: ignore
            d.consume(ExecutionCancelledEvent, None, self._execution_exited_cb),  # type: ignore
        ]

    async def aclose(self) -> None:
        for _evh in self._evhandlers:
            self._event_dispatcher.unconsume(_evh)
        await super().aclose()

    async def populate_config(self, raw_config: Mapping[str, Any]) -> None:
        config = self._config_iv.check(raw_config)
        self.idle_timeout = config['threshold']
        log.info(
            'TimeoutIdleChecker: default idle_timeout = {0:,} seconds',
            self.idle_timeout.total_seconds(),
        )

    async def update_app_streaming_status(
        self,
        session_id: SessionId,
        status: AppStreamingStatus,
    ) -> None:
        if status == AppStreamingStatus.HAS_ACTIVE_CONNECTIONS:
            await self._disable_timeout(session_id)
        elif status == AppStreamingStatus.NO_ACTIVE_CONNECTIONS:
            await self._update_timeout(session_id)

    async def _disable_timeout(self, session_id: SessionId) -> None:
        log.debug(f"TimeoutIdleChecker._disable_timeout({session_id})")
        await self._redis.set(f"session.{session_id}.last_access", "0", exist=self._redis.SET_IF_EXIST)

    async def _update_timeout(self, session_id: SessionId) -> None:
        log.debug(f"TimeoutIdleChecker._update_timeout({session_id})")
        t = await self._redis.time()
        await self._redis.set(
            f"session.{session_id}.last_access",
            f"{t:.06f}",
            expire=max(86400, self.idle_timeout.total_seconds() * 2),
        )

    async def _session_started_cb(
        self,
        context: None,
        source: AgentId,
        event: SessionStartedEvent,
    ) -> None:
        await self._update_timeout(event.session_id)

    async def _execution_started_cb(
        self,
        context: None,
        source: AgentId,
        event: ExecutionStartedEvent,
    ) -> None:
        await self._disable_timeout(event.session_id)

    async def _execution_exited_cb(
        self,
        context: None,
        source: AgentId,
        event: ExecutionFinishedEvent | ExecutionTimeoutEvent | ExecutionCancelledEvent,
    ) -> None:
        await self._update_timeout(event.session_id)

    async def _do_idle_check(
        self,
        context: None,
        source: AgentId,
        event: DoIdleCheckEvent,
    ) -> None:
        cache_token = self._policy_cache.set(dict())
        try:
            return await super()._do_idle_check(context, source, event)
        finally:
            self._policy_cache.reset(cache_token)

    async def check_session(self, session: RowProxy, dbconn: SAConnection) -> bool:
        session_id = session['id']
        active_streams = await self._redis.zcount(f"session.{session_id}.active_app_connections")
        if active_streams is not None and active_streams > 0:
            return True
        t = await self._redis.time()
        raw_last_access = await self._redis.get(f"session.{session_id}.last_access")
        if raw_last_access is None or raw_last_access == "0":
            return True
        last_access = float(raw_last_access)
        # serves as the default fallback if keypair resource policy's idle_timeout is "undefined"
        idle_timeout = self.idle_timeout.total_seconds()
        policy_cache = self._policy_cache.get()
        policy = policy_cache.get(session['access_key'], None)
        if policy is None:
            query = (
                sa.select([keypair_resource_policies])
                .select_from(
                    sa.join(
                        keypairs,
                        keypair_resource_policies,
                        (keypair_resource_policies.c.name == keypairs.c.resource_policy),
                    )
                )
                .where(
                    keypairs.c.access_key == session['access_key']
                )
            )
            result = await dbconn.execute(query)
            policy = await result.first()
            assert policy is not None
            policy_cache[session['access_key']] = policy
        # setting idle_timeout:
        # - zero/inf means "infinite"
        # - negative means "undefined"
        if policy['idle_timeout'] >= 0:
            idle_timeout = float(policy['idle_timeout'])
        if (
            (idle_timeout <= 0)
            or (math.isinf(idle_timeout) and idle_timeout > 0)
            or (t - last_access <= idle_timeout)
        ):
            return True
        return False


class UtilizationIdleChecker(BaseIdleChecker):
    """
    Checks the idleness of a session by the current utilization of all
    compute devices and agents assigned to it.
    """

    name: ClassVar[str] = "utilization"
    default_cpu_util_threshold: ClassVar[float] = 30.0
    default_accelerator_util_threshold: ClassVar[float] = 10.0
    default_initial_grace_period: ClassVar[float] = 300.0  # allow first 5 minutes to be idle

    async def populate_config(self, config: Mapping[str, Any]) -> None:
        pass

    async def check_session(self, session: RowProxy, dbconn: SAConnection) -> bool:
        # last_stat = session['last_stat']
        # TODO: implement
        return True

    async def update_app_streaming_status(
        self,
        session_id: SessionId,
        status: AppStreamingStatus,
    ) -> None:
        pass


checker_registry: Mapping[str, Type[BaseIdleChecker]] = {
    TimeoutIdleChecker.name: TimeoutIdleChecker,
    UtilizationIdleChecker.name: UtilizationIdleChecker,
}


async def create_idle_checkers(
    dbpool: SAPool,
    shared_config: SharedConfig,
    event_dispatcher: EventDispatcher,
    event_producer: EventProducer,
) -> Sequence[BaseIdleChecker]:
    """
    Create an instance of session idleness checker
    from the given configuration and using the given event dispatcher.
    """
    checkers = await shared_config.etcd.get('config/idle/enabled')
    if not checkers:
        return []
    instances = []
    for checker_name in checkers.split(","):
        checker_cls = checker_registry.get(checker_name, None)
        if checker_cls is None:
            log.warning("ignoring an unknown idle checker name: {checker_name}")
            continue
        log.info(f"Initializing idle checker: {checker_name}")
        checker_instance = await checker_cls.new(dbpool, shared_config, event_dispatcher, event_producer)
        instances.append(checker_instance)
    return instances



'''
The main web / websocket server
'''

import asyncio
from datetime import datetime
import functools
import importlib
import json
import logging
import os
import pwd, grp
import ssl
import sys
import traceback
from typing import (
    Any,
    AsyncIterator,
    Final,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    cast,
)

from aiohttp import web
import aiohttp_cors
import aiojobs.aiohttp
import aiotools
from aiopg.sa import create_engine
from aiopg.sa.engine import get_dialect
import click
from pathlib import Path
from setproctitle import setproctitle

from ai.backend.common import redis
from ai.backend.common.cli import LazyGroup
from ai.backend.common.events import EventDispatcher, EventProducer
from ai.backend.common.utils import env_info, current_loop
from ai.backend.common.json import ExtendedJSONEncoder
from ai.backend.common.logging import Logger, BraceStyleAdapter
from ai.backend.common.plugin.hook import HookPluginContext, ALL_COMPLETED, PASSED
from ai.backend.common.plugin.monitor import (
    ErrorPluginContext,
    StatsPluginContext,
    INCREMENT,
)

from ..manager import __version__
from ..manager.background import BackgroundTaskManager
from ..manager.exceptions import InvalidArgument
from ..manager.idle import create_idle_checkers
from ..manager.models.storage import StorageSessionManager
from ..manager.plugin.webapp import WebappPluginContext
from ..manager.registry import AgentRegistry
from ..manager.scheduler.dispatcher import SchedulerDispatcher
from .config import (
    LocalConfig,
    SharedConfig,
    load as load_config,
    volume_config_iv,
)
from .defs import REDIS_STAT_DB, REDIS_LIVE_DB, REDIS_IMAGE_DB, REDIS_STREAM_DB
from .exceptions import (
    BackendError,
    MethodNotAllowed,
    GenericNotFound,
    GenericBadRequest,
    InternalServerError,
    InvalidAPIParameters,
)
from .types import (
    AppCreator,
    WebRequestHandler, WebMiddleware,
    CleanupContext,
)
from . import ManagerStatus

VALID_VERSIONS: Final = frozenset([
    # 'v1.20160915',  # deprecated
    # 'v2.20170315',  # deprecated
    # 'v3.20170615',  # deprecated

    # authentication changed not to use request bodies
    'v4.20181215',

    # added & enabled streaming-execute API
    'v4.20190115',

    # changed resource/image formats
    'v4.20190315',

    # added user mgmt and ID/password authentication
    # added domain/group/scaling-group
    # added domain/group/scaling-group ref. fields to user/keypair/vfolder objects
    'v4.20190615',

    # added mount_map parameter when creating kernel
    # changed GraphQL query structures for multi-container bundled sessions
    'v5.20191215',

    # rewrote vfolder upload/download APIs to migrate to external storage proxies
    'v6.20200815',
])
LATEST_REV_DATES: Final = {
    1: '20160915',
    2: '20170915',
    3: '20181215',
    4: '20190615',
    5: '20191215',
    6: '20200815',
}
LATEST_API_VERSION: Final = 'v6.20200815'

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.server'))

PUBLIC_INTERFACES: Final = [
    'pidx',
    'background_task_manager',
    'local_config',
    'shared_config',
    'dbpool',
    'registry',
    'redis_live',
    'redis_stat',
    'redis_image',
    'redis_stream',
    'event_dispatcher',
    'event_producer',
    'idle_checkers',
    'storage_manager',
    'stats_monitor',
    'error_monitor',
    'hook_plugin_ctx',
]

public_interface_objs: MutableMapping[str, Any] = {}


async def hello(request: web.Request) -> web.Response:
    """
    Returns the API version number.
    """
    return web.json_response({
        'version': LATEST_API_VERSION,
        'manager': __version__,
    })


async def on_prepare(request: web.Request, response: web.StreamResponse) -> None:
    response.headers['Server'] = 'BackendAI'


@web.middleware
async def api_middleware(request: web.Request,
                         handler: WebRequestHandler) -> web.StreamResponse:
    _handler = handler
    method_override = request.headers.get('X-Method-Override', None)
    if method_override:
        request = request.clone(method=method_override)
        new_match_info = await request.app.router.resolve(request)
        if new_match_info is None:
            raise InternalServerError('No matching method handler found')
        _handler = new_match_info.handler
        request._match_info = new_match_info  # type: ignore  # this is a hack
    ex = request.match_info.http_exception
    if ex is not None:
        # handled by exception_middleware
        raise ex
    new_api_version = request.headers.get('X-BackendAI-Version')
    legacy_api_version = request.headers.get('X-Sorna-Version')
    api_version = new_api_version or legacy_api_version
    try:
        if api_version is None:
            path_major_version = int(request.match_info.get('version', 5))
            revision_date = LATEST_REV_DATES[path_major_version]
            request['api_version'] = (path_major_version, revision_date)
        elif api_version in VALID_VERSIONS:
            hdr_major_version, revision_date = api_version.split('.', maxsplit=1)
            request['api_version'] = (int(hdr_major_version[1:]), revision_date)
        else:
            return GenericBadRequest('Unsupported API version.')
    except (ValueError, KeyError):
        return GenericBadRequest('Unsupported API version.')
    resp = (await _handler(request))
    return resp


@web.middleware
async def exception_middleware(request: web.Request,
                               handler: WebRequestHandler) -> web.StreamResponse:
    app = request.app
    error_monitor = app['error_monitor']
    stats_monitor = app['stats_monitor']
    try:
        await stats_monitor.report_metric(INCREMENT, 'ai.backend.gateway.api.requests')
        resp = (await handler(request))
    except InvalidArgument as ex:
        if len(ex.args) > 1:
            raise InvalidAPIParameters(f"{ex.args[0]}: {', '.join(map(str, ex.args[1:]))}")
        elif len(ex.args) == 1:
            raise InvalidAPIParameters(ex.args[0])
        else:
            raise InvalidAPIParameters()
    except BackendError as ex:
        if ex.status_code == 500:
            log.warning('Internal server error raised inside handlers')
        await error_monitor.capture_exception()
        await stats_monitor.report_metric(INCREMENT, 'ai.backend.gateway.api.failures')
        await stats_monitor.report_metric(INCREMENT, f'ai.backend.gateway.api.status.{ex.status_code}')
        raise
    except web.HTTPException as ex:
        await stats_monitor.report_metric(INCREMENT, 'ai.backend.gateway.api.failures')
        await stats_monitor.report_metric(INCREMENT, f'ai.backend.gateway.api.status.{ex.status_code}')
        if ex.status_code == 404:
            raise GenericNotFound
        if ex.status_code == 405:
            concrete_ex = cast(web.HTTPMethodNotAllowed, ex)
            raise MethodNotAllowed(concrete_ex.method, concrete_ex.allowed_methods)
        log.warning('Bad request: {0!r}', ex)
        raise GenericBadRequest
    except asyncio.CancelledError as e:
        # The server is closing or the client has disconnected in the middle of
        # request.  Atomic requests are still executed to their ends.
        log.debug('Request cancelled ({0} {1})', request.method, request.rel_url)
        raise e
    except Exception as e:
        await error_monitor.capture_exception()
        log.exception('Uncaught exception in HTTP request handlers {0!r}', e)
        if app['local_config']['debug']['enabled']:
            raise InternalServerError(traceback.format_exc())
        else:
            raise InternalServerError()
    else:
        await stats_monitor.report_metric(INCREMENT, f'ai.backend.gateway.api.status.{resp.status}')
        return resp


@aiotools.actxmgr
async def shared_config_ctx(app: web.Application) -> AsyncIterator[None]:
    # populate public interfaces
    app['shared_config'] = SharedConfig(
        app,
        app['local_config']['etcd']['addr'],
        app['local_config']['etcd']['user'],
        app['local_config']['etcd']['password'],
        app['local_config']['etcd']['namespace'],
    )
    await app['shared_config'].reload()
    _update_public_interface_objs(app)
    yield
    await app['shared_config'].close()


@aiotools.actxmgr
async def webapp_plugin_ctx(root_app: web.Application) -> AsyncIterator[None]:
    ctx = WebappPluginContext(root_app['shared_config'].etcd, root_app['local_config'])
    await ctx.init()
    root_app['webapp_plugin_ctx'] = ctx
    for plugin_name, plugin_instance in ctx.plugins.items():
        if root_app['pidx'] == 0:
            log.info('Loading webapp plugin: {0}', plugin_name)
        subapp, global_middlewares = await plugin_instance.create_app(root_app['cors_opts'])
        _init_subapp(plugin_name, root_app, subapp, global_middlewares)
    yield
    await ctx.cleanup()


async def manager_status_ctx(app: web.Application) -> AsyncIterator[None]:
    if app['pidx'] == 0:
        mgr_status = await app['shared_config'].get_manager_status()
        if mgr_status is None or mgr_status not in (ManagerStatus.RUNNING, ManagerStatus.FROZEN):
            # legacy transition: we now have only RUNNING or FROZEN for HA setup.
            await app['shared_config'].update_manager_status(ManagerStatus.RUNNING)
            mgr_status = ManagerStatus.RUNNING
        log.info('Manager status: {}', mgr_status)
        tz = app['shared_config']['system']['timezone']
        log.info('Configured timezone: {}', tz.tzname(datetime.now()))
    yield


async def redis_ctx(app: web.Application) -> AsyncIterator[None]:
    app['redis_live'] = await redis.connect_with_retries(
        str(app['shared_config'].get_redis_url(db=REDIS_LIVE_DB)),
        timeout=3.0,
        encoding='utf8',
    )
    app['redis_stat'] = await redis.connect_with_retries(
        str(app['shared_config'].get_redis_url(db=REDIS_STAT_DB)),
        timeout=3.0,
        encoding='utf8',
    )
    app['redis_image'] = await redis.connect_with_retries(
        str(app['shared_config'].get_redis_url(db=REDIS_IMAGE_DB)),
        timeout=3.0,
        encoding='utf8',
    )
    app['redis_stream'] = await redis.connect_with_retries(
        str(app['shared_config'].get_redis_url(db=REDIS_STREAM_DB)),
        timeout=3.0,
        encoding='utf8',
    )
    _update_public_interface_objs(app)
    yield
    app['redis_image'].close()
    await app['redis_image'].wait_closed()
    app['redis_stat'].close()
    await app['redis_stat'].wait_closed()
    app['redis_live'].close()
    await app['redis_live'].wait_closed()
    app['redis_stream'].close()
    await app['redis_stream'].wait_closed()


async def database_ctx(app: web.Application) -> AsyncIterator[None]:
    app['dbpool'] = await create_engine(
        host=app['local_config']['db']['addr'].host, port=app['local_config']['db']['addr'].port,
        user=app['local_config']['db']['user'], password=app['local_config']['db']['password'],
        dbname=app['local_config']['db']['name'],
        echo=bool(app['local_config']['logging']['level'] == 'DEBUG'),
        minsize=8, maxsize=256,
        timeout=60, pool_recycle=120,
        dialect=get_dialect(
            json_serializer=functools.partial(json.dumps, cls=ExtendedJSONEncoder),
        ),
    )
    _update_public_interface_objs(app)
    yield
    app['dbpool'].close()
    await app['dbpool'].wait_closed()


async def event_dispatcher_ctx(app: web.Application) -> AsyncIterator[None]:
    local_config: LocalConfig = app['local_config']
    shared_config: SharedConfig = app['shared_config']

    async def redis_connector():
        redis_url = shared_config.get_redis_url(db=REDIS_STREAM_DB)
        return await redis.connect_with_retries(str(redis_url), encoding=None)

    app['event_producer'] = await EventProducer.new(redis_connector)
    app['event_dispatcher'] = await EventDispatcher.new(
        redis_connector,
        log_events=local_config['debug']['log-events'],
    )
    _update_public_interface_objs(app)
    yield
    await app['event_dispatcher'].close()
    await app['event_producer'].close()


async def idle_checker_ctx(app: web.Application) -> AsyncIterator[None]:
    app['idle_checkers'] = await create_idle_checkers(
        app['dbpool'], app['shared_config'], app['event_dispatcher'], app['event_producer'],
    )
    _update_public_interface_objs(app)
    yield
    for instance in app['idle_checkers']:
        await instance.aclose()


async def storage_manager_ctx(app: web.Application) -> AsyncIterator[None]:
    shared_config: SharedConfig = app['shared_config']
    raw_vol_config = await shared_config.etcd.get_prefix('volumes')
    config = volume_config_iv.check(raw_vol_config)
    app['storage_manager'] = StorageSessionManager(config)
    _update_public_interface_objs(app)
    yield
    await app['storage_manager'].aclose()


async def hook_plugin_ctx(app: web.Application) -> AsyncIterator[None]:
    ctx = HookPluginContext(app['shared_config'].etcd, app['local_config'])
    app['hook_plugin_ctx'] = ctx
    _update_public_interface_objs(app)
    await ctx.init()
    hook_result = await ctx.dispatch(
        'ACTIVATE_MANAGER',
        (),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RuntimeError('Could not activate the manager instance.')
    yield
    await ctx.cleanup()


async def agent_registry_ctx(app: web.Application) -> AsyncIterator[None]:
    app['registry'] = AgentRegistry(
        app['shared_config'],
        app['dbpool'],
        app['redis_stat'],
        app['redis_live'],
        app['redis_image'],
        app['event_dispatcher'],
        app['event_producer'],
        app['storage_manager'],
        app['hook_plugin_ctx'],
    )
    await app['registry'].init()
    _update_public_interface_objs(app)
    yield
    await app['registry'].shutdown()


async def sched_dispatcher_ctx(app: web.Application) -> AsyncIterator[None]:
    sched_dispatcher = await SchedulerDispatcher.new(
        app['local_config'], app['shared_config'],
        app['event_dispatcher'], app['event_producer'],
        app['registry'],
    )
    yield
    await sched_dispatcher.close()


async def monitoring_ctx(app: web.Application) -> AsyncIterator[None]:
    ectx = ErrorPluginContext(app['shared_config'].etcd, app['local_config'])
    sctx = StatsPluginContext(app['shared_config'].etcd, app['local_config'])
    await ectx.init(context={'app': app})
    await sctx.init()
    app['error_monitor'] = ectx
    app['stats_monitor'] = sctx
    _update_public_interface_objs(app)
    yield
    await sctx.cleanup()
    await ectx.cleanup()


async def background_task_ctx(app: web.Application) -> AsyncIterator[None]:
    app['background_task_manager'] = BackgroundTaskManager(app['event_producer'])
    _update_public_interface_objs(app)
    yield


async def background_task_ctx_shutdown(app: web.Application) -> None:
    if 'background_task_manager' in app:
        await app['background_task_manager'].shutdown()


setattr(background_task_ctx, 'shutdown', background_task_ctx_shutdown)


def handle_loop_error(
    root_app: web.Application,
    loop: asyncio.AbstractEventLoop,
    context: Mapping[str, Any],
) -> None:
    if isinstance(loop, aiojobs.Scheduler):
        loop = current_loop()
    exception = context.get('exception')
    msg = context.get('message', '(empty message)')
    if exception is not None:
        if sys.exc_info()[0] is not None:
            log.exception('Error inside event loop: {0}', msg)
            if 'error_monitor' in root_app:
                loop.create_task(root_app['error_monitor'].capture_exception())
        else:
            exc_info = (type(exception), exception, exception.__traceback__)
            log.error('Error inside event loop: {0}', msg, exc_info=exc_info)
            if 'error_monitor' in root_app:
                loop.create_task(root_app['error_monitor'].capture_exception(exc_instance=exception))


def _init_subapp(pkg_name: str,
                 root_app: web.Application,
                 subapp: web.Application,
                 global_middlewares: Iterable[WebMiddleware]) -> None:
    subapp.on_response_prepare.append(on_prepare)

    async def _copy_public_interface_objs(subapp: web.Application):
        # Allow subapp's access to the root app properties.
        # These are the public APIs exposed to plugins as well.
        for key, obj in public_interface_objs.items():
            subapp[key] = obj

    # We must copy the public interface prior to all user-defined startup signal handlers.
    subapp.on_startup.insert(0, _copy_public_interface_objs)
    prefix = subapp.get('prefix', pkg_name.split('.')[-1].replace('_', '-'))
    aiojobs.aiohttp.setup(subapp, **root_app['scheduler_opts'])
    root_app.add_subapp('/' + prefix, subapp)
    root_app.middlewares.extend(global_middlewares)


def init_subapp(pkg_name: str, root_app: web.Application, create_subapp: AppCreator) -> None:
    subapp, global_middlewares = create_subapp(root_app['cors_opts'])
    _init_subapp(pkg_name, root_app, subapp, global_middlewares)


def _update_public_interface_objs(root_app: web.Application) -> None:
    # This must be called in clean_ctx functions so that
    # we keep the public interface up-to-date.
    for key in PUBLIC_INTERFACES:
        if key in root_app and key not in public_interface_objs:
            public_interface_objs[key] = root_app[key]


def build_root_app(
    pidx: int,
    local_config: LocalConfig, *,
    cleanup_contexts: Sequence[CleanupContext] = None,
    subapp_pkgs: Sequence[str] = None,
    scheduler_opts: Mapping[str, Any] = None,
) -> web.Application:
    public_interface_objs.clear()
    app = web.Application(middlewares=[
        exception_middleware,
        api_middleware,
    ])
    global_exception_handler = functools.partial(handle_loop_error, app)
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(global_exception_handler)
    app['local_config'] = local_config
    app['cors_opts'] = {
        '*': aiohttp_cors.ResourceOptions(
            allow_credentials=False,
            expose_headers="*", allow_headers="*"),
    }
    default_scheduler_opts = {
        'limit': 2048,
        'close_timeout': 30,
        'exception_handler': global_exception_handler,
    }
    app['scheduler_opts'] = {
        **default_scheduler_opts,
        **(scheduler_opts if scheduler_opts is not None else {}),
    }
    app['pidx'] = pidx
    _update_public_interface_objs(app)
    app.on_response_prepare.append(on_prepare)

    if cleanup_contexts is None:
        cleanup_contexts = [
            manager_status_ctx,
            redis_ctx,
            database_ctx,
            event_dispatcher_ctx,
            idle_checker_ctx,
            storage_manager_ctx,
            hook_plugin_ctx,
            monitoring_ctx,
            agent_registry_ctx,
            sched_dispatcher_ctx,
            background_task_ctx,
        ]
    for cleanup_ctx in cleanup_contexts:
        if shutdown_cb := getattr(cleanup_ctx, 'shutdown', None):
            app.on_shutdown.append(shutdown_cb)
    app.cleanup_ctx.extend(cleanup_contexts)
    aiojobs.aiohttp.setup(app, **app['scheduler_opts'])
    cors = aiohttp_cors.setup(app, defaults=app['cors_opts'])
    # should be done in create_app() in other modules.
    cors.add(app.router.add_route('GET', r'', hello))
    cors.add(app.router.add_route('GET', r'/', hello))
    if subapp_pkgs is None:
        subapp_pkgs = []
    for pkg_name in subapp_pkgs:
        if pidx == 0:
            log.info('Loading module: {0}', pkg_name[1:])
        subapp_mod = importlib.import_module(pkg_name, 'ai.backend.gateway')
        init_subapp(pkg_name, app, getattr(subapp_mod, 'create_app'))
    return app


@aiotools.actxmgr
async def server_main(loop: asyncio.AbstractEventLoop,
                      pidx: int,
                      _args: List[Any]) -> AsyncIterator[None]:
    subapp_pkgs = [
        '.etcd', '.events',
        '.auth', '.ratelimit',
        '.vfolder', '.admin',
        '.session',
        '.stream',
        '.manager',
        '.resource',
        '.scaling_group',
        '.cluster_template',
        '.session_template',
        '.image',
        '.userconfig',
        '.domainconfig',
        '.groupconfig',
        '.logs',
    ]
    app = build_root_app(pidx, _args[0], subapp_pkgs=subapp_pkgs)

    # Plugin webapps should be loaded before runner.setup(),
    # which freezes on_startup event.
    async with shared_config_ctx(app), \
               webapp_plugin_ctx(app):
        ssl_ctx = None
        if app['local_config']['manager']['ssl-enabled']:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(app['local_config']['manager']['ssl-cert']),
                str(app['local_config']['manager']['ssl-privkey']),
            )
        runner = web.AppRunner(app)
        await runner.setup()
        service_addr = app['local_config']['manager']['service-addr']
        site = web.TCPSite(
            runner,
            str(service_addr.host),
            service_addr.port,
            backlog=1024,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        await site.start()

        if os.geteuid() == 0:
            uid = app['local_config']['manager']['user']
            gid = app['local_config']['manager']['group']
            os.setgroups([
                g.gr_gid for g in grp.getgrall()
                if pwd.getpwuid(uid).pw_name in g.gr_mem
            ])
            os.setgid(gid)
            os.setuid(uid)
            log.info('changed process uid and gid to {}:{}', uid, gid)
        log.info('started handling API requests at {}', service_addr)

        try:
            yield
        finally:
            log.info('shutting down...')
            await runner.cleanup()


@aiotools.actxmgr
async def server_main_logwrapper(loop: asyncio.AbstractEventLoop,
                                 pidx: int, _args: List[Any]) -> AsyncIterator[None]:
    setproctitle(f"backend.ai: manager worker-{pidx}")
    log_endpoint = _args[1]
    logger = Logger(_args[0]['logging'], is_master=False, log_endpoint=log_endpoint)
    try:
        with logger:
            async with server_main(loop, pidx, _args):
                yield
    except Exception:
        traceback.print_exc()


@click.group(invoke_without_command=True)
@click.option('-f', '--config-path', '--config', type=Path, default=None,
              help='The config file path. (default: ./manager.toml and /etc/backend.ai/manager.toml)')
@click.option('--debug', is_flag=True,
              help='Enable the debug mode and override the global log level to DEBUG.')
@click.pass_context
def main(ctx: click.Context, config_path: Path, debug: bool) -> None:

    cfg = load_config(config_path, debug)

    if ctx.invoked_subcommand is None:
        cfg['manager']['pid-file'].write_text(str(os.getpid()))
        log_sockpath = Path(f'/tmp/backend.ai/ipc/manager-logger-{os.getpid()}.sock')
        log_sockpath.parent.mkdir(parents=True, exist_ok=True)
        log_endpoint = f'ipc://{log_sockpath}'
        try:
            logger = Logger(cfg['logging'], is_master=True, log_endpoint=log_endpoint)
            with logger:
                ns = cfg['etcd']['namespace']
                setproctitle(f"backend.ai: manager {ns}")
                log.info('Backend.AI Gateway {0}', __version__)
                log.info('runtime: {0}', env_info())
                log_config = logging.getLogger('ai.backend.gateway.config')
                log_config.debug('debug mode enabled.')

                if cfg['manager']['event-loop'] == 'uvloop':
                    import uvloop
                    uvloop.install()
                    log.info('Using uvloop as the event loop backend')
                try:
                    aiotools.start_server(
                        server_main_logwrapper,
                        num_workers=cfg['manager']['num-proc'],
                        args=(cfg, log_endpoint),
                    )
                finally:
                    log.info('terminated.')
        finally:
            if cfg['manager']['pid-file'].is_file():
                # check is_file() to prevent deleting /dev/null!
                cfg['manager']['pid-file'].unlink()
    else:
        # Click is going to invoke a subcommand.
        pass


@main.group(cls=LazyGroup, import_name='ai.backend.gateway.auth:cli')
def auth() -> None:
    pass


if __name__ == '__main__':
    sys.exit(main())



"""enumerate_vfolder_perms

Revision ID: a1fd4e7b7782
Revises: f9971fbb34d9
Create Date: 2018-09-05 16:51:49.973195

"""
from alembic import op
import sqlalchemy as sa
from ai.backend.manager.models import VFolderPermission
from sqlalchemy.dialects import postgresql


# revision identifiers, used by Alembic.
revision = 'a1fd4e7b7782'
down_revision = 'f9971fbb34d9'
branch_labels = None
depends_on = None

# NOTE: VFolderPermission is EnumValueType
vfperm_choices = list(map(lambda v: v.value, VFolderPermission))
vfolderpermission = postgresql.ENUM(
    *vfperm_choices,
    name='vfolderpermission',
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    vfolderpermission.create(op.get_bind())
    op.alter_column('vfolder_invitations', column_name='permission',
                    type_=sa.Enum(*vfperm_choices, name='vfolderpermission'),
                    postgresql_using='permission::vfolderpermission')
    op.alter_column('vfolder_permissions', column_name='permission',
                    type_=sa.Enum(*vfperm_choices, name='vfolderpermission'),
                    postgresql_using='permission::vfolderpermission')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('vfolder_invitations', column_name='permission',
                    type_=sa.String(length=2),
                    postgresql_using='permission::text::vfolderpermission')
    op.alter_column('vfolder_permissions', column_name='permission',
                    type_=sa.String(length=2),
                    postgresql_using='permission::text::vfolderpermission')
    vfolderpermission.drop(op.get_bind())
    # ### end Alembic commands ###



from collections import ChainMap
from datetime import datetime, timedelta
import functools
import hashlib, hmac
import logging
import secrets
from typing import (
    Any, Final,
    Iterable,
    Tuple,
)

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
import click
from dateutil.tz import tzutc
from dateutil.parser import parse as dtparse
import sqlalchemy as sa
import trafaret as t

from ai.backend.common import validators as tx
from ai.backend.common.logging import Logger, BraceStyleAdapter
from ai.backend.common.plugin.hook import (
    ALL_COMPLETED,
    FIRST_COMPLETED,
    PASSED,
)
from .exceptions import (
    AuthorizationFailed,
    GenericBadRequest,
    GenericForbidden,
    GenericNotFound,
    InternalServerError,
    InvalidAuthParameters,
    InvalidAPIParameters,
    RejectedByHook,
)
from ..manager.models import (
    keypairs, keypair_resource_policies, users,
)
from ..manager.models.user import UserRole, UserStatus, INACTIVE_USER_STATUSES, check_credential
from ..manager.models.keypair import generate_keypair as _gen_keypair, generate_ssh_keypair
from ..manager.models.group import association_groups_users, groups
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params, set_handler_attr, get_handler_attr

log: Final = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.auth'))

whois_timezone_info: Final = {
    "A": 1 * 3600,
    "ACDT": 10.5 * 3600,
    "ACST": 9.5 * 3600,
    "ACT": -5 * 3600,
    "ACWST": 8.75 * 3600,
    "ADT": 4 * 3600,
    "AEDT": 11 * 3600,
    "AEST": 10 * 3600,
    "AET": 10 * 3600,
    "AFT": 4.5 * 3600,
    "AKDT": -8 * 3600,
    "AKST": -9 * 3600,
    "ALMT": 6 * 3600,
    "AMST": -3 * 3600,
    "AMT": -4 * 3600,
    "ANAST": 12 * 3600,
    "ANAT": 12 * 3600,
    "AQTT": 5 * 3600,
    "ART": -3 * 3600,
    "AST": 3 * 3600,
    "AT": -4 * 3600,
    "AWDT": 9 * 3600,
    "AWST": 8 * 3600,
    "AZOST": 0 * 3600,
    "AZOT": -1 * 3600,
    "AZST": 5 * 3600,
    "AZT": 4 * 3600,
    "AoE": -12 * 3600,
    "B": 2 * 3600,
    "BNT": 8 * 3600,
    "BOT": -4 * 3600,
    "BRST": -2 * 3600,
    "BRT": -3 * 3600,
    "BST": 6 * 3600,
    "BTT": 6 * 3600,
    "C": 3 * 3600,
    "CAST": 8 * 3600,
    "CAT": 2 * 3600,
    "CCT": 6.5 * 3600,
    "CDT": -5 * 3600,
    "CEST": 2 * 3600,
    "CET": 1 * 3600,
    "CHADT": 13.75 * 3600,
    "CHAST": 12.75 * 3600,
    "CHOST": 9 * 3600,
    "CHOT": 8 * 3600,
    "CHUT": 10 * 3600,
    "CIDST": -4 * 3600,
    "CIST": -5 * 3600,
    "CKT": -10 * 3600,
    "CLST": -3 * 3600,
    "CLT": -4 * 3600,
    "COT": -5 * 3600,
    "CST": -6 * 3600,
    "CT": -6 * 3600,
    "CVT": -1 * 3600,
    "CXT": 7 * 3600,
    "ChST": 10 * 3600,
    "D": 4 * 3600,
    "DAVT": 7 * 3600,
    "DDUT": 10 * 3600,
    "E": 5 * 3600,
    "EASST": -5 * 3600,
    "EAST": -6 * 3600,
    "EAT": 3 * 3600,
    "ECT": -5 * 3600,
    "EDT": -4 * 3600,
    "EEST": 3 * 3600,
    "EET": 2 * 3600,
    "EGST": 0 * 3600,
    "EGT": -1 * 3600,
    "EST": -5 * 3600,
    "ET": -5 * 3600,
    "F": 6 * 3600,
    "FET": 3 * 3600,
    "FJST": 13 * 3600,
    "FJT": 12 * 3600,
    "FKST": -3 * 3600,
    "FKT": -4 * 3600,
    "FNT": -2 * 3600,
    "G": 7 * 3600,
    "GALT": -6 * 3600,
    "GAMT": -9 * 3600,
    "GET": 4 * 3600,
    "GFT": -3 * 3600,
    "GILT": 12 * 3600,
    "GMT": 0 * 3600,
    "GST": 4 * 3600,
    "GYT": -4 * 3600,
    "H": 8 * 3600,
    "HDT": -9 * 3600,
    "HKT": 8 * 3600,
    "HOVST": 8 * 3600,
    "HOVT": 7 * 3600,
    "HST": -10 * 3600,
    "I": 9 * 3600,
    "ICT": 7 * 3600,
    "IDT": 3 * 3600,
    "IOT": 6 * 3600,
    "IRDT": 4.5 * 3600,
    "IRKST": 9 * 3600,
    "IRKT": 8 * 3600,
    "IRST": 3.5 * 3600,
    "IST": 5.5 * 3600,
    "JST": 9 * 3600,
    "K": 10 * 3600,
    "KGT": 6 * 3600,
    "KOST": 11 * 3600,
    "KRAST": 8 * 3600,
    "KRAT": 7 * 3600,
    "KST": 9 * 3600,
    "KUYT": 4 * 3600,
    "L": 11 * 3600,
    "LHDT": 11 * 3600,
    "LHST": 10.5 * 3600,
    "LINT": 14 * 3600,
    "M": 12 * 3600,
    "MAGST": 12 * 3600,
    "MAGT": 11 * 3600,
    "MART": 9.5 * 3600,
    "MAWT": 5 * 3600,
    "MDT": -6 * 3600,
    "MHT": 12 * 3600,
    "MMT": 6.5 * 3600,
    "MSD": 4 * 3600,
    "MSK": 3 * 3600,
    "MST": -7 * 3600,
    "MT": -7 * 3600,
    "MUT": 4 * 3600,
    "MVT": 5 * 3600,
    "MYT": 8 * 3600,
    "N": -1 * 3600,
    "NCT": 11 * 3600,
    "NDT": 2.5 * 3600,
    "NFT": 11 * 3600,
    "NOVST": 7 * 3600,
    "NOVT": 7 * 3600,
    "NPT": 5.5 * 3600,
    "NRT": 12 * 3600,
    "NST": 3.5 * 3600,
    "NUT": -11 * 3600,
    "NZDT": 13 * 3600,
    "NZST": 12 * 3600,
    "O": -2 * 3600,
    "OMSST": 7 * 3600,
    "OMST": 6 * 3600,
    "ORAT": 5 * 3600,
    "P": -3 * 3600,
    "PDT": -7 * 3600,
    "PET": -5 * 3600,
    "PETST": 12 * 3600,
    "PETT": 12 * 3600,
    "PGT": 10 * 3600,
    "PHOT": 13 * 3600,
    "PHT": 8 * 3600,
    "PKT": 5 * 3600,
    "PMDT": -2 * 3600,
    "PMST": -3 * 3600,
    "PONT": 11 * 3600,
    "PST": -8 * 3600,
    "PT": -8 * 3600,
    "PWT": 9 * 3600,
    "PYST": -3 * 3600,
    "PYT": -4 * 3600,
    "Q": -4 * 3600,
    "QYZT": 6 * 3600,
    "R": -5 * 3600,
    "RET": 4 * 3600,
    "ROTT": -3 * 3600,
    "S": -6 * 3600,
    "SAKT": 11 * 3600,
    "SAMT": 4 * 3600,
    "SAST": 2 * 3600,
    "SBT": 11 * 3600,
    "SCT": 4 * 3600,
    "SGT": 8 * 3600,
    "SRET": 11 * 3600,
    "SRT": -3 * 3600,
    "SST": -11 * 3600,
    "SYOT": 3 * 3600,
    "T": -7 * 3600,
    "TAHT": -10 * 3600,
    "TFT": 5 * 3600,
    "TJT": 5 * 3600,
    "TKT": 13 * 3600,
    "TLT": 9 * 3600,
    "TMT": 5 * 3600,
    "TOST": 14 * 3600,
    "TOT": 13 * 3600,
    "TRT": 3 * 3600,
    "TVT": 12 * 3600,
    "U": -8 * 3600,
    "ULAST": 9 * 3600,
    "ULAT": 8 * 3600,
    "UTC": 0 * 3600,
    "UYST": -2 * 3600,
    "UYT": -3 * 3600,
    "UZT": 5 * 3600,
    "V": -9 * 3600,
    "VET": -4 * 3600,
    "VLAST": 11 * 3600,
    "VLAT": 10 * 3600,
    "VOST": 6 * 3600,
    "VUT": 11 * 3600,
    "W": -10 * 3600,
    "WAKT": 12 * 3600,
    "WARST": -3 * 3600,
    "WAST": 2 * 3600,
    "WAT": 1 * 3600,
    "WEST": 1 * 3600,
    "WET": 0 * 3600,
    "WFT": 12 * 3600,
    "WGST": -2 * 3600,
    "WGT": -3 * 3600,
    "WIB": 7 * 3600,
    "WIT": 9 * 3600,
    "WITA": 8 * 3600,
    "WST": 14 * 3600,
    "WT": 0 * 3600,
    "X": -11 * 3600,
    "Y": -12 * 3600,
    "YAKST": 10 * 3600,
    "YAKT": 9 * 3600,
    "YAPT": 10 * 3600,
    "YEKST": 6 * 3600,
    "YEKT": 5 * 3600,
    "Z": 0 * 3600,
}


def _extract_auth_params(request):
    """
    HTTP Authorization header must be formatted as:
    "Authorization: BackendAI signMethod=HMAC-SHA256,
                    credential=<ACCESS_KEY>:<SIGNATURE>"
    """
    auth_hdr = request.headers.get('Authorization')
    if not auth_hdr:
        return None
    pieces = auth_hdr.split(' ', 1)
    if len(pieces) != 2:
        raise InvalidAuthParameters('Malformed authorization header')
    auth_type, auth_str = pieces
    if auth_type not in ('BackendAI', 'Sorna'):
        raise InvalidAuthParameters('Invalid authorization type name')

    raw_params = map(lambda s: s.strip(), auth_str.split(','))
    params = {}
    for param in raw_params:
        key, value = param.split('=', 1)
        params[key.strip()] = value.strip()

    try:
        access_key, signature = params['credential'].split(':', 1)
        ret = params['signMethod'], access_key, signature
        return ret
    except (KeyError, ValueError):
        raise InvalidAuthParameters('Missing or malformed authorization parameters')


def check_date(request) -> bool:
    raw_date = request.headers.get('Date')
    if not raw_date:
        raw_date = request.headers.get('X-BackendAI-Date',
                                       request.headers.get('X-Sorna-Date'))
    if not raw_date:
        return False
    try:
        # HTTP standard says "Date" header must be in GMT only.
        # However, dateutil.parser can recognize other commonly used
        # timezone names and offsets.
        date = dtparse(raw_date, tzinfos=whois_timezone_info)
        if date.tzinfo is None:
            date = date.replace(tzinfo=tzutc())  # assume as UTC
        now = datetime.now(tzutc())
        min_date = now - timedelta(minutes=15)
        max_date = now + timedelta(minutes=15)
        request['date'] = date
        request['raw_date'] = raw_date
        if not (min_date < date < max_date):
            return False
    except ValueError:
        return False
    return True


async def sign_request(sign_method, request, secret_key) -> str:
    try:
        mac_type, hash_type = map(lambda s: s.lower(), sign_method.split('-'))
        assert mac_type == 'hmac', 'Unsupported request signing method (MAC type)'
        assert hash_type in hashlib.algorithms_guaranteed, \
               'Unsupported request signing method (hash type)'

        new_api_version = request.headers.get('X-BackendAI-Version')
        legacy_api_version = request.headers.get('X-Sorna-Version')
        api_version = new_api_version or legacy_api_version
        assert api_version is not None, 'API version missing in request headers'
        body = b''
        if api_version < 'v4.20181215':
            if (request.can_read_body and
                request.content_type != 'multipart/form-data'):
                # read the whole body if neither streaming nor bodyless
                body = await request.read()
        body_hash = hashlib.new(hash_type, body).hexdigest()

        sign_bytes = ('{0}\n{1}\n{2}\nhost:{3}\ncontent-type:{4}\n'
                      'x-{name}-version:{5}\n{6}').format(
            request.method, str(request.raw_path), request['raw_date'],
            request.host, request.content_type, api_version,
            body_hash,
            name='backendai' if new_api_version is not None else 'sorna'
        ).encode()
        sign_key = hmac.new(secret_key.encode(),
                            request['date'].strftime('%Y%m%d').encode(),
                            hash_type).digest()
        sign_key = hmac.new(sign_key, request.host.encode(), hash_type).digest()
        return hmac.new(sign_key, sign_bytes, hash_type).hexdigest()
    except ValueError:
        raise AuthorizationFailed('Invalid signature')
    except AssertionError as e:
        raise InvalidAuthParameters(e.args[0])


@web.middleware
async def auth_middleware(request, handler):
    '''
    Fetches user information and sets up keypair, uesr, and is_authorized
    attributes.
    '''
    # This is a global middleware: request.app is the root app.
    request['is_authorized'] = False
    request['is_admin'] = False
    request['is_superadmin'] = False
    request['keypair'] = None
    request['user'] = None
    if not get_handler_attr(request, 'auth_required', False):
        return (await handler(request))
    if not check_date(request):
        raise InvalidAuthParameters('Date/time sync error')
    params = _extract_auth_params(request)
    if params:
        sign_method, access_key, signature = params
        async with request.app['dbpool'].acquire() as conn, conn.begin():
            j = (keypairs.join(users, keypairs.c.user == users.c.uuid)
                         .join(keypair_resource_policies,
                               keypairs.c.resource_policy == keypair_resource_policies.c.name))
            query = (sa.select([users, keypairs, keypair_resource_policies], use_labels=True)
                       .select_from(j)
                       .where((keypairs.c.access_key == access_key) &
                              (keypairs.c.is_active.is_(True))))
            result = await conn.execute(query)
            row = await result.first()
            if row is None:
                raise AuthorizationFailed('Access key not found')
            my_signature = \
                await sign_request(sign_method, request, row['keypairs_secret_key'])
            if secrets.compare_digest(my_signature, signature):
                query = (keypairs.update()
                                 .values(last_used=datetime.now(tzutc()),
                                         num_queries=keypairs.c.num_queries + 1)
                                 .where(keypairs.c.access_key == access_key))
                await conn.execute(query)
                request['is_authorized'] = True
                request['keypair'] = {
                    col.name: row[f'keypairs_{col.name}']
                    for col in keypairs.c
                    if col.name != 'secret_key'
                }
                request['keypair']['resource_policy'] = {
                    col.name: row[f'keypair_resource_policies_{col.name}']
                    for col in keypair_resource_policies.c
                }
                request['user'] = {
                    col.name: row[f'users_{col.name}']
                    for col in users.c
                    if col.name not in ('password', 'description', 'created_at')
                }
                request['user']['id'] = row['keypairs_user_id']  # legacy
                # if request['role'] in ['admin', 'superadmin']:
                if row['keypairs_is_admin']:
                    request['is_admin'] = True
                if request['user']['role'] == 'superadmin':
                    request['is_superadmin'] = True

    # No matter if authenticated or not, pass-through to the handler.
    # (if it's required, auth_required decorator will handle the situation.)
    return (await handler(request))


def auth_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


def admin_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False) and request.get('is_admin', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


def superadmin_required(handler):

    @functools.wraps(handler)
    async def wrapped(request, *args, **kwargs):
        if request.get('is_authorized', False) and request.get('is_superadmin', False):
            return (await handler(request, *args, **kwargs))
        raise AuthorizationFailed('Unauthorized access')

    set_handler_attr(wrapped, 'auth_required', True)
    return wrapped


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('echo'): t.String,
    }))
async def test(request: web.Request, params: Any) -> web.Response:
    log.info('AUTH.TEST(ak:{})', request['keypair']['access_key'])
    resp_data = {'authorized': 'yes'}
    if 'echo' in params:
        resp_data['echo'] = params['echo']
    return web.json_response(resp_data)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('group', default=None): t.Null | tx.UUID,
    }))
async def get_role(request: web.Request, params: Any) -> web.Response:
    group_role = None
    log.info('AUTH.ROLES(ak:{}, d:{}, g:{})',
             request['keypair']['access_key'],
             request['user']['domain_name'],
             params['group'])
    if params['group'] is not None:
        query = (
            # TODO: per-group role is not yet implemented.
            sa.select([association_groups_users.c.group_id])
            .select_from(association_groups_users)
            .where(
                (association_groups_users.c.group_id == params['group']) &
                (association_groups_users.c.user_id == request['user']['uuid'])
            )
        )
        async with request.app['dbpool'].acquire() as conn:
            result = await conn.execute(query)
            row = await result.first()
            if row is None:
                raise GenericNotFound('No such user group or '
                                      'you are not the member of the group.')
        group_role = 'user'
    resp_data = {
        'global_role': 'superadmin' if request['is_superadmin'] else 'user',
        'domain_role': 'admin' if request['is_admin'] else 'user',
        'group_role': group_role,
    }
    return web.json_response(resp_data)


@atomic
@check_api_params(
    t.Dict({
        t.Key('type'): t.Enum('keypair', 'jwt'),
        t.Key('domain'): t.String,
        t.Key('username'): t.String,
        t.Key('password'): t.String,
    }))
async def authorize(request: web.Request, params: Any) -> web.Response:
    if params['type'] != 'keypair':
        # other types are not implemented yet.
        raise InvalidAPIParameters('Unsupported authorization type')
    log.info('AUTH.AUTHORIZE(d:{0[domain]}, u:{0[username]}, passwd:****, type:{0[type]})', params)
    dbpool = request.app['dbpool']

    # [Hooking point for AUTHORIZE with the FIRST_COMPLETED requirement]
    # The hook handlers should accept the whole ``params`` dict, and optional
    # ``dbpool`` parameter (if the hook needs to query to database).
    # They should return a corresponding Backend.AI user object after performing
    # their own authentication steps, like LDAP authentication, etc.
    hook_result = await request.app['hook_plugin_ctx'].dispatch(
        'AUTHORIZE',
        (params, dbpool),
        return_when=FIRST_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RejectedByHook.from_hook_result(hook_result)
    elif hook_result.result:
        # Passed one of AUTHORIZED hook
        user = hook_result.result
    else:
        # No AUTHORIZE hook is defined (proceed with normal login)
        user = await check_credential(
            dbpool,
            params['domain'], params['username'], params['password']
        )
    if user is None:
        raise AuthorizationFailed('User credential mismatch.')
    if user.get('status') == UserStatus.BEFORE_VERIFICATION:
        raise AuthorizationFailed('This account needs email verification.')
    if user.get('status') in INACTIVE_USER_STATUSES:
        raise AuthorizationFailed('User credential mismatch.')
    async with dbpool.acquire() as conn:
        query = (sa.select([keypairs.c.access_key, keypairs.c.secret_key])
                   .select_from(keypairs)
                   .where(
                       (keypairs.c.user == user['uuid']) &
                       (keypairs.c.is_active)
                   )
                   .order_by(sa.desc(keypairs.c.is_admin)))
        result = await conn.execute(query)
        keypair = await result.first()
    if keypair is None:
        raise AuthorizationFailed('No API keypairs found.')
    return web.json_response({
        'data': {
            'access_key': keypair['access_key'],
            'secret_key': keypair['secret_key'],
            'role': user['role'],
            'status': user['status'],
        },
    })


@atomic
@check_api_params(
    t.Dict({
        t.Key('domain'): t.String,
        t.Key('email'): t.String,
        t.Key('password'): t.String,
    }).allow_extra('*'))
async def signup(request: web.Request, params: Any) -> web.Response:
    log_fmt = 'AUTH.SIGNUP(d:{}, email:{}, passwd:****)'
    log_args = (params['domain'], params['email'])
    log.info(log_fmt, *log_args)
    dbpool = request.app['dbpool']

    # [Hooking point for PRE_SIGNUP with the ALL_COMPLETED requirement]
    # The hook handlers should accept the whole ``params`` dict.
    # They should return a dict to override the user information,
    # where the keys must be a valid field name of the users table,
    # with two exceptions: "resource_policy" (name) and "group" (name).
    # A plugin may return an empty dict if it has nothing to override.
    hook_result = await request.app['hook_plugin_ctx'].dispatch(
        'PRE_SIGNUP',
        (params, ),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        raise RejectedByHook.from_hook_result(hook_result)
    else:
        # Merge the hook results as a single map.
        user_data_overriden = ChainMap(*hook_result.result)

    async with dbpool.acquire() as conn:
        # Check if email already exists.
        query = (sa.select([users])
                   .select_from(users)
                   .where((users.c.email == params['email'])))
        result = await conn.execute(query)
        row = await result.first()
        if row is not None:
            raise GenericBadRequest('Email already exists')

        # Create a user.
        data = {
            'domain_name': params['domain'],
            'username': params['username'] if 'username' in params else params['email'],
            'email': params['email'],
            'password': params['password'],
            'need_password_change': False,
            'full_name': params['full_name'] if 'full_name' in params else '',
            'description': params['description'] if 'description' in params else '',
            'status': UserStatus.ACTIVE,
            'status_info': 'user-signup',
            'role': UserRole.USER,
            'integration_id': None,
        }
        if user_data_overriden:
            for key, val in user_data_overriden.items():
                if key in data:  # take only valid fields
                    data[key] = val
        query = (users.insert().values(data))
        result = await conn.execute(query)
        if result.rowcount > 0:
            checkq = users.select().where(users.c.email == params['email'])
            result = await conn.execute(checkq)
            user = await result.first()
            # Create user's first access_key and secret_key.
            ak, sk = _gen_keypair()
            resource_policy = (
                user_data_overriden.get('resource_policy', 'default')
            )
            kp_data = {
                'user_id': params['email'],
                'access_key': ak,
                'secret_key': sk,
                'is_active': True if data.get('status') == UserStatus.ACTIVE else False,
                'is_admin': False,
                'resource_policy': resource_policy,
                'concurrency_used': 0,
                'rate_limit': 1000,
                'num_queries': 0,
                'user': user.uuid,
            }
            query = (keypairs.insert().values(kp_data))
            await conn.execute(query)

            # Add user to the default group.
            group_name = user_data_overriden.get('group', 'default')
            query = (sa.select([groups.c.id])
                       .select_from(groups)
                       .where(groups.c.domain_name == params['domain'])
                       .where(groups.c.name == group_name))
            result = await conn.execute(query)
            grp = await result.first()
            if grp is not None:
                values = [{'user_id': user.uuid, 'group_id': grp.id}]
                query = association_groups_users.insert().values(values)
                await conn.execute(query)
        else:
            raise InternalServerError('Error creating user account')

    resp_data = {
        'access_key': ak,
        'secret_key': sk,
    }

    # [Hooking point for POST_SIGNUP as one-way notification]
    # The hook handlers should accept a tuple of the user email,
    # the new user's UUID, and a dict with initial user's preferences.
    initial_user_prefs = {
        'lang': request.headers.get('Accept-Language', 'en-us').split(',')[0].lower(),
    }
    await request.app['hook_plugin_ctx'].notify(
        'POST_SIGNUP',
        (params['email'], user.uuid, initial_user_prefs)
    )
    return web.json_response(resp_data, status=201)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(['email', 'username']): t.String,
        t.Key('password'): t.String,
    }))
async def signout(request: web.Request, params: Any) -> web.Response:
    domain_name = request['user']['domain_name']
    log.info('AUTH.SIGNOUT(d:{}, email:{})', domain_name, params['email'])
    dbpool = request.app['dbpool']
    if request['user']['email'] != params['email']:
        raise GenericForbidden('Not the account owner')
    result = await check_credential(
        dbpool,
        domain_name, params['email'], params['password'])
    if result is None:
        raise GenericBadRequest('Invalid email and/or password')
    async with dbpool.acquire() as conn, conn.begin():
        # Inactivate the user.
        query = (users.update()
                      .values(status=UserStatus.INACTIVE)
                      .where(users.c.email == params['email']))
        await conn.execute(query)
        # Inactivate every keypairs of the user.
        query = (keypairs.update()
                         .values(is_active=False)
                         .where(keypairs.c.user_id == params['email']))
        await conn.execute(query)
    return web.json_response({})


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('old_password'): t.String,
        t.Key('new_password'): t.String,
        t.Key('new_password2'): t.String,
    }))
async def update_password(request: web.Request, params: Any) -> web.Response:
    domain_name = request['user']['domain_name']
    email = request['user']['email']
    log_fmt = 'AUTH.UDPATE_PASSWORD(d:{}, email:{})'
    log_args = (domain_name, email)
    log.info(log_fmt, *log_args)
    dbpool = request.app['dbpool']

    user = await check_credential(dbpool, domain_name, email, params['old_password'])
    if user is None:
        log.info(log_fmt + ': old password mismtach', *log_args)
        raise AuthorizationFailed('Old password mismatch')
    if params['new_password'] != params['new_password2']:
        log.info(log_fmt + ': new password mismtach', *log_args)
        return web.json_response({'error_msg': 'new password mismitch'}, status=400)

    # [Hooking point for VERIFY_PASSWORD_FORMAT with the ALL_COMPLETED requirement]
    # The hook handlers should accept the old password and the new password and implement their
    # own password validation rules.
    # They should return None if the validation is successful and raise the Reject error
    # otherwise.
    hook_result = await request.app['hook_plugin_ctx'].dispatch(
        'VERIFY_PASSWORD_FORMAT',
        (params['old_password'], params['new_password']),
        return_when=ALL_COMPLETED,
    )
    if hook_result.status != PASSED:
        hook_result.reason = hook_result.reason or 'invalid password format'
        raise RejectedByHook.from_hook_result(hook_result)

    async with dbpool.acquire() as conn:
        # Update user password.
        data = {
            'password': params['new_password'],
            'need_password_change': False,
        }
        query = (users.update().values(data).where(users.c.email == email))
        await conn.execute(query)
    return web.json_response({}, status=200)


@atomic
@auth_required
async def get_ssh_keypair(request: web.Request) -> web.Response:
    domain_name = request['user']['domain_name']
    access_key = request['keypair']['access_key']
    log_fmt = 'AUTH.GET_SSH_KEYPAIR(d:{}, ak:{})'
    log_args = (domain_name, access_key)
    log.info(log_fmt, *log_args)
    dbpool = request.app['dbpool']
    async with dbpool.acquire() as conn:
        # Get SSH public key. Return partial string from the public key just for checking.
        query = (sa.select([keypairs.c.ssh_public_key])
                   .where(keypairs.c.access_key == access_key))
        pubkey = await conn.scalar(query)
    return web.json_response({'ssh_public_key': pubkey}, status=200)


@atomic
@auth_required
async def refresh_ssh_keypair(request: web.Request) -> web.Response:
    domain_name = request['user']['domain_name']
    access_key = request['keypair']['access_key']
    log_fmt = 'AUTH.REFRESH_SSH_KEYPAIR(d:{}, ak:{})'
    log_args = (domain_name, access_key)
    log.info(log_fmt, *log_args)
    dbpool = request.app['dbpool']
    async with dbpool.acquire() as conn:
        pubkey, privkey = generate_ssh_keypair()
        data = {
            'ssh_public_key': pubkey,
            'ssh_private_key': privkey,
        }
        query = (keypairs.update()
                         .values(data)
                         .where(keypairs.c.access_key == access_key))
        await conn.execute(query)
    return web.json_response(data, status=200)


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app['prefix'] = 'auth'  # slashed to distinguish with "/vN/authorize"
    app['api_versions'] = (1, 2, 3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    root_resource = cors.add(app.router.add_resource(r''))
    cors.add(root_resource.add_route('GET', test))
    cors.add(root_resource.add_route('POST', test))
    test_resource = cors.add(app.router.add_resource('/test'))
    cors.add(test_resource.add_route('GET', test))
    cors.add(test_resource.add_route('POST', test))
    cors.add(app.router.add_route('POST', '/authorize', authorize))
    cors.add(app.router.add_route('GET', '/role', get_role))
    cors.add(app.router.add_route('POST', '/signup', signup))
    cors.add(app.router.add_route('POST', '/signout', signout))
    cors.add(app.router.add_route('POST', '/update-password', update_password))
    cors.add(app.router.add_route('GET', '/ssh-keypair', get_ssh_keypair))
    cors.add(app.router.add_route('PATCH', '/ssh-keypair', refresh_ssh_keypair))
    return app, [auth_middleware]


@click.group()
def cli():
    pass


@cli.command()
def generate_keypair():
    logger = Logger({
        'level': 'INFO',
        'drivers': ['console'],
        'pkg-ns': {'ai.backend': 'INFO'},
        'console': {'colored': True, 'format': 'verbose'},
    })
    with logger:
        log.info('generating keypair...')
        ak, sk = _gen_keypair()
        print(f'Access Key: {ak} ({len(ak)} bytes)')
        print(f'Secret Key: {sk} ({len(sk)} bytes)')


if __name__ == '__main__':
    cli()



from __future__ import annotations

from typing import (
    Awaitable, Callable, Iterable,
    AsyncIterator,
    Tuple,
    Mapping,
)

from aiohttp import web
import aiohttp_cors


WebRequestHandler = Callable[
    [web.Request],
    Awaitable[web.StreamResponse]
]
WebMiddleware = Callable[
    [web.Request, WebRequestHandler],
    Awaitable[web.StreamResponse]
]

CORSOptions = Mapping[str, aiohttp_cors.ResourceOptions]
AppCreator = Callable[
    [CORSOptions],
    Tuple[web.Application, Iterable[WebMiddleware]]
]

CleanupContext = Callable[[web.Application], AsyncIterator[None]]



"""
Common definitions/constants used throughout the API gateway.
"""

from typing import Final

# Redis database IDs depending on purposes
REDIS_STAT_DB: Final = 0
REDIS_RLIM_DB: Final = 1
REDIS_LIVE_DB: Final = 2
REDIS_IMAGE_DB: Final = 3
REDIS_STREAM_DB: Final = 4



import enum
from typing import (
    Protocol,
)

import sqlalchemy as sa
from aiopg.sa.connection import SAConnection


class SessionGetter(Protocol):

    def __call__(self, *, db_connection: SAConnection) -> sa.engine.RowProxy:
        ...


# Sentinel is a special object that indicates a special status instead of a value
# where the user expects a value.
# According to the discussion in https://github.com/python/typing/issues/236,
# we define our Sentinel type as an enum with only one special value.
# This enables passing of type checks by "value is sentinel" (or "value is Sentinel.token")
# instead of more expensive "isinstance(value, Sentinel)" because we can assure type checkers
# to think there is no other possible instances of the Sentinel type.

class Sentinel(enum.Enum):
    token = 0



"""convert-lang-to-image-and-registry

Revision ID: 2b0931e4a059
Revises: f0f4ee907155
Create Date: 2019-01-28 23:53:44.342786

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2b0931e4a059'
down_revision = 'f0f4ee907155'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', column_name='lang', new_column_name='image')
    op.add_column('kernels', sa.Column('registry', sa.String(length=512),
                  nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('kernels', column_name='image', new_column_name='lang')
    op.drop_column('kernels', 'registry')
    # ### end Alembic commands ###



"""
This module defines a series of Backend.AI's plugin-specific errors.
"""
from aiohttp import web
from ai.backend.gateway.exceptions import BackendError


class PluginError(web.HTTPBadRequest, BackendError):
    error_type  = 'https://api.backend.ai/probs/plugin-error'
    error_title = 'Plugin generated error'


class PluginConfigurationError(PluginError):
    error_type  = 'https://api.backend.ai/probs/plugin-config-error'
    error_title = 'Plugin configuration error'



import base64
import secrets
from typing import (
    Any,
    Iterable,
    Tuple,
)

from aiohttp import web
import aiohttp_cors
import jinja2
import sqlalchemy as sa
import trafaret as t

from ai.backend.common import validators as tx
from ai.backend.common.docker import ImageRef
from ai.backend.common.etcd import (
    quote as etcd_quote,
)
from ai.backend.common.types import (
    SessionTypes,
)

from .auth import admin_required
from .exceptions import InvalidAPIParameters
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from ..manager.defs import DEFAULT_ROLE
from ..manager.models import (
    domains, groups, query_allowed_sgroups,
    association_groups_users as agus,
)
from .types import CORSOptions, WebMiddleware
from .utils import (
    check_api_params,
)


DOCKERFILE_TEMPLATE = r'''# syntax = docker/dockerfile:1.0-experimental
FROM {{ src }}
MAINTAINER Backend.AI Manager

USER root

{% if runtime_type == 'python' -%}
ENV PYTHONUNBUFFERED=1 \
    LANG=C.UTF-8

RUN --mount=type=bind,source=wheelhouse,target=/root/wheelhouse \
    PIP_OPTS="--no-cache-dir --no-index --find-links=/root/wheelhouse" && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} -U pip setuptools && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} Pillow && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} h5py && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} ipython && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} jupyter && \
    {{ runtime_path }} -m pip install ${PIP_OPTS} jupyterlab

# Install ipython kernelspec
RUN {{ runtime_path }} -m ipykernel install \
    --prefix={{ runtime_path.parent.parent }} \
    --display-name "{{ brand }} on Backend.AI"
{%- endif %}

LABEL ai.backend.kernelspec="1" \
      ai.backend.envs.corecount="{{ cpucount_envvars | join(',') }}" \
      ai.backend.features="{% if has_ipykernel %}query batch {% endif %}uid-match" \
      ai.backend.resource.min.cpu="{{ min_cpu }}" \
      ai.backend.resource.min.mem="{{ min_mem }}" \
      ai.backend.resource.preferred.shmem="{{ pref_shmem }}" \
      ai.backend.accelerators="{{ accelerators | join(',') }}" \
{%- if 'cuda' is in accelerators %}
      ai.backend.resource.min.cuda.device=1 \
      ai.backend.resource.min.cuda.shares=0.1 \
{%- endif %}
      ai.backend.base-distro="{{ base_distro }}" \
{%- if service_ports %}
      ai.backend.service-ports="{% for item in service_ports -%}
          {{- item['name'] }}:
          {{- item['protocol'] }}:
          {%- if (item['ports'] | length) > 1 -%}
              [{{ item['ports'] | join(',') }}]
          {%- else -%}
              {{ item['ports'][0] }}
          {%- endif -%}
          {{- ',' if not loop.last }}
      {%- endfor %}" \
{%- endif %}
      ai.backend.runtime-type="{{ runtime_type }}" \
      ai.backend.runtime-path="{{ runtime_path }}"
'''  # noqa


@server_status_required(READ_ALLOWED)
@admin_required
async def get_import_image_form(request: web.Request) -> web.Response:
    async with request.app['dbpool'].acquire() as conn, conn.begin():
        query = (
            sa.select([groups.c.name])
            .select_from(
                sa.join(groups, domains,
                        groups.c.domain_name == domains.c.name)
            )
            .where(
                (domains.c.name == request['user']['domain_name']) &
                (domains.c.is_active) &
                (groups.c.is_active)
            )
        )
        result = await conn.execute(query)
        rows = await result.fetchall()
        accessible_groups = [row['name'] for row in rows]

        # FIXME: Currently this only consider domain-level scaling group associations,
        #        thus ignoring the group name query.
        rows = await query_allowed_sgroups(
            conn, request['user']['domain_name'], '', request['keypair']['access_key'])
        accessible_scaling_groups = [row['name'] for row in rows]

    return web.json_response({
        'fieldGroups': [
            {
                'name': 'Import options',
                'fields': [
                    {
                        'name': 'src',
                        'type': 'string',
                        'label': 'Source Docker image',
                        'placeholder': 'index.docker.io/lablup/tensorflow:2.0-source',
                        'help': 'The full Docker image name to import from. '
                                'The registry must be accessible by the client.',
                    },
                    {
                        'name': 'target',
                        'type': 'string',
                        'label': 'Target Docker image',
                        'placeholder': 'index.docker.io/lablup/tensorflow:2.0-target',
                        'help': 'The full Docker image name of the imported image.'
                                'The registry must be accessible by the client.',
                    },
                    {
                        'name': 'brand',
                        'type': 'string',
                        'label': 'Name of Jupyter kernel',
                        'placeholder': 'TensorFlow 2.0',
                        'help': 'The name of kernel to be shown in the Jupyter\'s kernel menu. '
                                'This will be suffixed with "on Backend.AI".',
                    },
                    {
                        'name': 'baseDistro',
                        'type': 'choice',
                        'choices': ['ubuntu', 'centos'],
                        'default': 'ubuntu',
                        'label': 'Base LINUX distribution',
                        'help': 'The base Linux distribution used by the source image',
                    },
                    {
                        'name': 'minCPU',
                        'type': 'number',
                        'min': 1,
                        'max': None,
                        'label': 'Minimum required CPU core(s)',
                        'help': 'The minimum number of CPU cores required by the image',
                    },
                    {
                        'name': 'minMemory',
                        'type': 'binarysize',
                        'min': '64m',
                        'max': None,
                        'label': 'Minimum required memory size',
                        'help': 'The minimum size of the main memory required by the image',
                    },
                    {
                        'name': 'preferredSharedMemory',
                        'type': 'binarysize',
                        'min': '64m',
                        'max': None,
                        'label': 'Preferred shared memory size',
                        'help': 'The preferred (default) size of the shared memory',
                    },
                    {
                        'name': 'supportedAccelerators',
                        'type': 'multichoice[str]',
                        'choices': ['cuda'],
                        'default': 'cuda',
                        'label': 'Supported accelerators',
                        'help': 'The list of accelerators supported by the image',
                    },
                    {
                        'name': 'runtimeType',
                        'type': 'choice',
                        'choices': ['python'],
                        'default': 'python',
                        'label': 'Runtime type of the image',
                        'help': 'The runtime type of the image. '
                                'Currently, the source image must have installed Python 2.7, 3.5, 3.6, '
                                'or 3.7 at least to import. '
                                'This will be used as the kernel of Jupyter service in this image.',
                    },
                    {
                        'name': 'runtimePath',
                        'type': 'string',
                        'default': '/usr/local/bin/python',
                        'label': 'Path of the runtime',
                        'placeholder': '/usr/local/bin/python',
                        'help': 'The path to the main executalbe of runtime language of the image. '
                                'Even for the same "python"-based images, this may differ significantly '
                                'image by image. (e.g., /usr/bin/python, /usr/local/bin/python, '
                                '/opt/something/bin/python, ...) '
                                'Please check this carefully not to get confused with OS-default ones '
                                'and custom-installed ones.',
                    },
                    {
                        'name': 'CPUCountEnvs',
                        'type': 'list[string]',
                        'default': ['NPROC', 'OMP_NUM_THREADS', 'OPENBLAS_NUM_THREADS'],
                        'label': 'CPU count environment variables',
                        'help': 'The name of environment variables to be overriden to the number of CPU '
                                'cores actually allocated to the container. Required for legacy '
                                'computation libraries.',
                    },
                    {
                        'name': 'servicePorts',
                        'type': 'multichoice[template]',
                        'templates': [
                            {'name': 'jupyter', 'protocol': 'http', 'ports': [8080]},
                            {'name': 'jupyterlab', 'protocol': 'http', 'ports': [8090]},
                            {'name': 'tensorboard', 'protocol': 'http', 'ports': [6006]},
                            {'name': 'digits', 'protocol': 'http', 'ports': [5000]},
                            {'name': 'vscode', 'protocol': 'http', 'ports': [8180]},
                            {'name': 'h2o-dai', 'protocol': 'http', 'ports': [12345]},
                        ],
                        'label': 'Supported service ports',
                        'help': 'The list of service ports supported by this image. '
                                'Note that sshd (port 2200) and ttyd (port 7681) are intrinsic; '
                                'they are always included regardless of the source image. '
                                'The port number 2000-2003 are reserved by Backend.AI, and '
                                'all port numbers must be larger than 1024 and smaller than 65535.',
                    },
                ]
            },
            {
                'name': 'Import Task Options',
                'help': 'The import task uses 1 CPU core and 2 GiB of memory.',
                'fields': [
                    {
                        'name': 'group',
                        'type': 'choice',
                        'choices': accessible_groups,
                        'label': 'Group to build image',
                        'help': 'The user group where the import task will be executed.',
                    },
                    {
                        'name': 'scalingGroup',
                        'type': 'choice',
                        'choices': accessible_scaling_groups,
                        'label': 'Scaling group to build image',
                        'help': 'The scaling group where the import task will take resources from.',
                    },
                ]
            },
        ]
    })


@server_status_required(ALL_ALLOWED)
@admin_required
@check_api_params(
    t.Dict({
        t.Key('src'): t.String,
        t.Key('target'): t.String,
        t.Key('launchOptions', default={}): t.Dict({
            t.Key('scalingGroup', default='default'): t.String,
            t.Key('group', default='default'): t.String,
        }).allow_extra('*'),
        t.Key('brand'): t.String,
        t.Key('baseDistro'): t.Enum('ubuntu', 'centos'),
        t.Key('minCPU', default=1): t.Int[1:],
        t.Key('minMemory', default='64m'): tx.BinarySize,
        t.Key('preferredSharedMemory', default='64m'): tx.BinarySize,
        t.Key('supportedAccelerators'): t.List(t.String),
        t.Key('runtimeType'): t.Enum('python'),
        t.Key('runtimePath'): tx.Path(type='file', allow_nonexisting=True, resolve=False),
        t.Key('CPUCountEnvs'): t.List(t.String),
        t.Key('servicePorts', default=[]): t.List(t.Dict({
            t.Key('name'): t.String,
            t.Key('protocol'): t.Enum('http', 'tcp', 'pty'),
            t.Key('ports'): t.List(t.Int[1:65535], min_length=1),
        })),
    }).allow_extra('*'))
async def import_image(request: web.Request, params: Any) -> web.Response:
    '''
    Import a docker image and convert it to a Backend.AI-compatible one,
    by automatically installing a few packages and adding image labels.

    Currently we only support auto-conversion of Python-based kernels (e.g.,
    NGC images) which has its own Python version installed.

    Internally, it launches a temporary kernel in an arbitrary agent within
    the client's domain, the "default" group, and the "default" scaling group.
    (The client may change the group and scaling group using *launchOptions.*
    If the client is a super-admin, it uses the "default" domain.)

    This temporary kernel occupies only 1 CPU core and 1 GiB memory.
    The kernel concurrency limit is not applied here, but we choose an agent
    based on their resource availability.
    The owner of this kernel is always the client that makes the API request.

    This API returns immediately after launching the temporary kernel.
    The client may check the progress of the import task using session logs.
    '''

    tpl = jinja2.Template(DOCKERFILE_TEMPLATE)

    async with request.app['dbpool'].acquire() as conn, conn.begin():
        query = (
            sa.select([domains.c.allowed_docker_registries])
            .select_from(domains)
            .where(domains.c.name == request['user']['domain_name'])
        )
        result = await conn.execute(query)
        allowed_docker_registries = await result.scalar()

    source_image = ImageRef(params['src'], allowed_docker_registries)
    target_image = ImageRef(params['target'], allowed_docker_registries)

    # TODO: validate and convert arguments to template variables
    dockerfile_content = tpl.render({
        'base_distro': params['baseDistro'],
        'cpucount_envvars': ['NPROC', 'OMP_NUM_THREADS', 'OPENBLAS_NUM_THREADS'],
        'runtime_type': params['runtimeType'],
        'runtime_path': params['runtimePath'],
        'service_ports': params['servicePorts'],
        'min_cpu': params['minCPU'],
        'min_mem': params['minMemory'],
        'pref_shmem': params['preferredSharedMemory'],
        'accelerators': params['supportedAccelerators'],
        'src': params['src'],
        'brand': params['brand'],
        'has_ipykernel': True,  # TODO: in the future, we may allow import of service-port only kernels.
    })

    sess_id = f'image-import-{secrets.token_urlsafe(8)}'
    access_key = request['keypair']['access_key']
    registry = request.app['registry']
    resource_policy = request['keypair']['resource_policy']

    async with request.app['dbpool'].acquire() as conn, conn.begin():
        query = (
            sa.select([groups.c.id])
            .select_from(
                sa.join(groups, domains,
                        groups.c.domain_name == domains.c.name)
            )
            .where(
                (domains.c.name == request['user']['domain_name']) &
                (groups.c.name == params['launchOptions']['group']) &
                (domains.c.is_active) &
                (groups.c.is_active)
            )
        )
        result = await conn.execute(query)
        group_id = await result.scalar()
        if group_id is None:
            raise InvalidAPIParameters("Invalid domain or group.")

        query = (
            sa.select([agus])
            .select_from(agus)
            .where(
                (agus.c.user_id == request['user']['uuid']) &
                (agus.c.group_id == group_id)
            )
        )
        result = await conn.execute(query)
        row = await result.first()
        if row is None:
            raise InvalidAPIParameters("You do not belong to the given group.")

    importer_image = ImageRef(
        request.app['local_config']['manager']['importer-image'],
        allowed_docker_registries,
    )

    docker_creds = {}
    for img_ref in (source_image, target_image):
        registry_info = await request.app['shared_config'].etcd.get_prefix_dict(
            f'config/docker/registry/{etcd_quote(img_ref.registry)}')
        docker_creds[img_ref.registry] = {
            'username': registry_info.get('username'),
            'password': registry_info.get('password'),
        }

    kernel_id = await registry.enqueue_session(
        sess_id, access_key,
        [{
            'image_ref': importer_image,
            'cluster_role': DEFAULT_ROLE,
            'creation_config': {
                'resources': {'cpu': '1', 'mem': '2g'},
                'scaling_group': params['launchOptions']['scalingGroup'],
                'environ': {
                    'SRC_IMAGE': source_image.canonical,
                    'TARGET_IMAGE': target_image.canonical,
                    'RUNTIME_PATH': params['runtimePath'],
                    'BUILD_SCRIPT': (
                        base64.b64encode(dockerfile_content.encode('utf8')).decode('ascii')
                    ),
                }
            },
            'startup_command': '/root/build-image.sh',
            'bootstrap_script': ''
        }],
        None,
        SessionTypes.BATCH,
        resource_policy,
        domain_name=request['user']['domain_name'],
        group_id=group_id,
        user_uuid=request['user']['uuid'],
        user_role=request['user']['role'],
        internal_data={
            'domain_socket_proxies': ['/var/run/docker.sock'],
            'docker_credentials': docker_creds,
            'prevent_vfolder_mounts': True,
            'block_service_ports': True,
        }
    )
    return web.json_response({
        'importTask': {
            'sessionId': sess_id,
            'taskId': str(kernel_id),
        },
    }, status=200)


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['prefix'] = 'image'
    app['api_versions'] = (4,)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('GET', '/import', get_import_image_form))
    cors.add(app.router.add_route('POST', '/import', import_image))
    return app, []



import datetime as dt
from datetime import datetime
import functools
import logging
import uuid

from aiohttp import web
import aiohttp_cors
import aioredlock
import aiotools
import sqlalchemy as sa
import trafaret as t
from typing import Any, Tuple, MutableMapping

from ai.backend.common import validators as tx
from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common.types import LogSeverity

from .auth import auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

from ..manager.models import (
    error_logs, UserRole, groups,
    association_groups_users as agus
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.logs'))


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(t.Dict(
    {
        t.Key('severity'): tx.Enum(LogSeverity),
        t.Key('source'): t.String,
        t.Key('message'): t.String,
        t.Key('context_lang'): t.String,
        t.Key('context_env'): tx.JSONString,
        t.Key('request_url', default=None): t.Null | t.String,
        t.Key('request_status', default=None): t.Null | t.Int,
        t.Key('traceback', default=None): t.Null | t.String
    }
))
async def append(request: web.Request, params: Any) -> web.Response:
    params['domain'] = request['user']['domain_name']
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    requester_uuid = request['user']['uuid']
    log.info('CREATE (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')

    dbpool = request.app['dbpool']

    async with dbpool.acquire() as conn, conn.begin():
        resp = {
            'success': True
        }
        query = error_logs.insert().values({
            'severity': params['severity'],
            'source': params['source'],
            'user': requester_uuid,
            'message': params['message'],
            'context_lang': params['context_lang'],
            'context_env': params['context_env'],
            'request_url': params['request_url'],
            'request_status': params['request_status'],
            'traceback': params['traceback'],
        })
        result = await conn.execute(query)
        assert result.rowcount == 1
    return web.json_response(resp)


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('mark_read', default=False): t.ToBool(),
        t.Key('page_size', default=20): t.ToInt(lt=101),
        t.Key('page_no', default=1): t.ToInt()
    }),
)
async def list_logs(request: web.Request, params: Any) -> web.Response:
    resp: MutableMapping[str, Any] = {'logs': []}
    dbpool = request.app['dbpool']
    domain_name = request['user']['domain_name']
    user_role = request['user']['role']
    user_uuid = request['user']['uuid']

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info('LIST (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')
    async with dbpool.acquire() as conn:
        is_admin = True
        query = (sa.select('*')
                   .select_from(error_logs)
                   .order_by(sa.desc(error_logs.c.created_at))
                   .limit(params['page_size']))
        count_query = (sa.select([sa.func.count(error_logs.c.message)])
                         .select_from(error_logs))
        if params['page_no'] > 1:
            query = query.offset((params['page_no'] - 1) * params['page_size'])
        if request['is_superadmin']:
            pass
        elif user_role == UserRole.ADMIN or user_role == 'admin':
            j = (groups.join(agus, groups.c.id == agus.c.group_id))
            usr_query = (sa.select([agus.c.user_id])
                           .select_from(j)
                           .where([groups.c.domain_name == domain_name]))
            result = await conn.execute(usr_query)
            usrs = await result.fetchall()
            user_ids = [g.id for g in usrs]
            where = error_logs.c.user.in_(user_ids)
            query = query.where(where)
            count_query = query.where(where)
        else:
            is_admin = False
            where = ((error_logs.c.user == user_uuid) &
                     (~error_logs.c.is_cleared))
            query = query.where(where)
            count_query = query.where(where)

        result = await conn.execute(query)
        async for row in result:
            result_item = {
                'log_id': str(row['id']),
                'created_at': datetime.timestamp(row['created_at']),
                'severity': row['severity'],
                'source': row['source'],
                'user': row['user'],
                'is_read': row['is_read'],
                'message': row['message'],
                'context_lang': row['context_lang'],
                'context_env': row['context_env'],
                'request_url': row['request_url'],
                'request_status': row['request_status'],
                'traceback': row['traceback'],
            }
            if result_item['user'] is not None:
                result_item['user'] = str(result_item['user'])
            if is_admin:
                result_item['is_cleared'] = row['is_cleared']
            resp['logs'].append(result_item)
        resp['count'] = await conn.scalar(count_query)
        if params['mark_read']:
            update = (sa.update(error_logs)
                        .values(is_read=True)
                        .where(error_logs.c.id.in_([x['log_id'] for x in resp['logs']])))
            await conn.execute(update)
        return web.json_response(resp, status=200)


@auth_required
@server_status_required(READ_ALLOWED)
async def mark_cleared(request: web.Request) -> web.Response:
    dbpool = request.app['dbpool']
    domain_name = request['user']['domain_name']
    user_role = request['user']['role']
    user_uuid = request['user']['uuid']
    log_id = uuid.UUID(request.match_info['log_id'])

    log.info('CLEAR')
    async with dbpool.acquire() as conn, conn.begin():
        query = (sa.update(error_logs)
                   .values(is_cleared=True))
        if request['is_superadmin']:
            query = query.where(error_logs.c.id == log_id)
        elif user_role == UserRole.ADMIN or user_role == 'admin':
            j = (groups.join(agus, groups.c.id == agus.c.group_id))
            usr_query = (sa.select([agus.c.user_id])
                           .select_from(j)
                           .where([groups.c.domain_name == domain_name]))
            result = await conn.execute(usr_query)
            usrs = await result.fetchall()
            user_ids = [g.id for g in usrs]
            query = query.where((error_logs.c.user.in_(user_ids)) &
                                (error_logs.c.id == log_id))
        else:
            query = (query.where((error_logs.c.user == user_uuid) &
                                 (error_logs.c.id == log_id)))

        result = await conn.execute(query)
        assert result.rowcount == 1

        return web.json_response({'success': True}, status=200)


async def log_cleanup_task(app: web.Application, interval):
    dbpool = app['dbpool']
    etcd = app['shared_config'].etcd
    raw_lifetime = await etcd.get('config/logs/error/retention')
    if raw_lifetime is None:
        raw_lifetime = '90d'
    checker = tx.TimeDuration()
    try:
        lifetime = checker.check(raw_lifetime)
    except ValueError:
        lifetime = dt.timedelta(days=90)
        log.info('Retention value specified in etcd not recognized by'
                 'trafaret validator, falling back to 90 days')
    boundary = datetime.now() - lifetime
    try:
        lock = await app['log_cleanup_lock'].lock('gateway.logs')
        async with lock:
            async with dbpool.acquire() as conn, conn.begin():
                query = (sa.select([error_logs.c.id])
                            .select_from(error_logs)
                            .where(error_logs.c.created_at < boundary))
                result = await conn.execute(query)
                log_ids = []
                async for row in result:
                    log_ids.append(row['id'])
                if len(log_ids) > 0:
                    log.info('Cleaning up {} log{}', len(log_ids), 's' if len(log_ids) > 1 else '')
                query = error_logs.delete().where(error_logs.c.id.in_(log_ids))
                result = await conn.execute(query)
                assert result.rowcount == len(log_ids)

    except aioredlock.LockError:
        log.debug('schedule(): temporary locking failure; will be retried.')


async def init(app: web.Application) -> None:
    redis_url = app['shared_config'].get_redis_url()
    app['log_cleanup_lock'] = aioredlock.Aioredlock([str(redis_url)])
    app['log_cleanup_task'] = aiotools.create_timer(
        functools.partial(log_cleanup_task, app), 5.0)


async def shutdown(app: web.Application) -> None:
    app['log_cleanup_task'].cancel()
    await app['log_cleanup_task']


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['api_versions'] = (4, 5)
    app['prefix'] = 'logs/error'
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('POST', '', append))
    cors.add(app.router.add_route('GET', '', list_logs))
    cors.add(app.router.add_route('POST', r'/{log_id}/clear', mark_cleared))

    return app, []



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager as actxmgr
from contextvars import ContextVar
import itertools
from typing import (
    Any,
    AsyncIterator,
    Final,
    Iterable,
    List,
    Mapping,
    Sequence,
    Tuple,
    TypedDict,
    TYPE_CHECKING,
)
from uuid import UUID

import aiohttp
import attr
import graphene
import yarl

from .base import (
    Item, PaginatedList,
)
from ..exceptions import InvalidArgument
if TYPE_CHECKING:
    from ..registry import AgentRegistry

__all__ = (
    'StorageProxyInfo',
    'VolumeInfo',
    'StorageSessionManager',
    'StorageVolume',
)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class StorageProxyInfo:
    session: aiohttp.ClientSession
    secret: str
    client_api_url: yarl.URL
    manager_api_url: yarl.URL


AUTH_TOKEN_HDR: Final = 'X-BackendAI-Storage-Auth-Token'

_ctx_volumes_cache: ContextVar[List[Tuple[str, VolumeInfo]]] = ContextVar('_ctx_volumes')


class VolumeInfo(TypedDict):
    name: str
    backend: str
    path: str
    fsprefix: str
    capabilities: List[str]


class StorageSessionManager:

    _proxies: Mapping[str, StorageProxyInfo]

    def __init__(self, storage_config: Mapping[str, Any]) -> None:
        self.config = storage_config
        self._proxies = {}
        for proxy_name, proxy_config in self.config['proxies'].items():
            connector = aiohttp.TCPConnector(ssl=proxy_config['ssl_verify'])
            session = aiohttp.ClientSession(connector=connector)
            self._proxies[proxy_name] = StorageProxyInfo(
                session=session,
                secret=proxy_config['secret'],
                client_api_url=yarl.URL(proxy_config['client_api']),
                manager_api_url=yarl.URL(proxy_config['manager_api']),
            )

    async def aclose(self) -> None:
        close_aws = []
        for proxy_info in self._proxies.values():
            close_aws.append(proxy_info.session.close())
        await asyncio.gather(*close_aws, return_exceptions=True)

    @staticmethod
    def split_host(vfolder_host: str) -> Tuple[str, str]:
        proxy_name, _, volume_name = vfolder_host.partition(':')
        return proxy_name, volume_name

    async def get_all_volumes(self) -> Iterable[Tuple[str, VolumeInfo]]:
        try:
            # per-asyncio-task cache
            return _ctx_volumes_cache.get()
        except LookupError:
            pass
        fetch_aws = []

        async def _fetch(
            proxy_name: str,
            proxy_info: StorageProxyInfo,
        ) -> Iterable[Tuple[str, VolumeInfo]]:
            async with proxy_info.session.request(
                'GET', proxy_info.manager_api_url / 'volumes',
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return ((proxy_name, volume_data) for volume_data in reply['volumes'])

        for proxy_name, proxy_info in self._proxies.items():
            fetch_aws.append(_fetch(proxy_name, proxy_info))
        results = [*itertools.chain(*await asyncio.gather(*fetch_aws))]
        _ctx_volumes_cache.set(results)
        return results

    async def get_mount_path(self, vfolder_host: str, vfolder_id: UUID) -> str:
        async with self.request(
            vfolder_host, 'GET', 'folder/mount',
            json={
                'volume': self.split_host(vfolder_host)[1],
                'vfid': str(vfolder_id),
            },
        ) as (_, resp):
            reply = await resp.json()
            return reply['path']

    @actxmgr
    async def request(
        self,
        vfolder_host_or_proxy_name: str,
        method: str,
        request_relpath: str,
        /,
        *args,
        **kwargs,
    ) -> AsyncIterator[Tuple[yarl.URL, aiohttp.ClientResponse]]:
        proxy_name, _ = self.split_host(vfolder_host_or_proxy_name)
        try:
            proxy_info = self._proxies[proxy_name]
        except KeyError:
            raise InvalidArgument('There is no such storage proxy', proxy_name)
        headers = kwargs.pop('headers', {})
        headers[AUTH_TOKEN_HDR] = proxy_info.secret
        async with proxy_info.session.request(
            method, proxy_info.manager_api_url / request_relpath,
            *args,
            headers=headers,
            **kwargs,
        ) as client_resp:
            yield proxy_info.client_api_url, client_resp


class StorageVolume(graphene.ObjectType):

    class Meta:
        interfaces = (Item, )

    # id: {proxy_name}:{name}
    backend = graphene.String()
    fsprefix = graphene.String()
    path = graphene.String()
    capabilities = graphene.List(graphene.String)
    hardware_metadata = graphene.JSONString()
    performance_metric = graphene.JSONString()
    usage = graphene.JSONString()

    async def resolve_hardware_metadata(self, info: graphene.ResolveInfo):
        registry: AgentRegistry = info.context['registry']
        return await registry.gather_storage_hwinfo(self.id)

    async def resolve_performance_metric(self, info: graphene.ResolveInfo):
        storage_manager: StorageSessionManager = info.context['storage_manager']
        proxy_name, volume_name = storage_manager.split_host(self.id)
        try:
            proxy_info = storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        try:
            async with proxy_info.session.request(
                'GET', proxy_info.manager_api_url / 'volume/performance-metric',
                json={'volume': volume_name},
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return reply['metric']
        except aiohttp.ClientResponseError:
            return {}

    async def resolve_usage(self, info: graphene.ResolveInfo):
        storage_manager: StorageSessionManager = info.context['storage_manager']
        proxy_name, volume_name = storage_manager.split_host(self.id)
        try:
            proxy_info = storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        try:
            async with proxy_info.session.request(
                'GET', proxy_info.manager_api_url / 'folder/fs-usage',
                json={'volume': volume_name},
                raise_for_status=True,
                headers={AUTH_TOKEN_HDR: proxy_info.secret},
            ) as resp:
                reply = await resp.json()
                return reply
        except aiohttp.ClientResponseError:
            return {}

    @classmethod
    def from_info(cls, proxy_name: str, volume_info: VolumeInfo) -> StorageVolume:
        return cls(
            id=f"{proxy_name}:{volume_info['name']}",
            backend=volume_info['backend'],
            path=volume_info['path'],
            fsprefix=volume_info['fsprefix'],
            capabilities=volume_info['capabilities'],
        )

    @classmethod
    async def load_count(
        cls,
        context,
    ) -> int:
        storage_manager: StorageSessionManager = context['storage_manager']
        volumes = [*await storage_manager.get_all_volumes()]
        return len(volumes)

    @classmethod
    async def load_slice(
        cls, context, limit: int, offset: int,
    ) -> Sequence[StorageVolume]:
        storage_manager: StorageSessionManager = context['storage_manager']
        volumes = [*await storage_manager.get_all_volumes()]
        return [
            cls.from_info(proxy_name, volume_info)
            for proxy_name, volume_info in volumes[offset:offset + limit]
        ]

    @classmethod
    async def load_by_id(
        cls, context, id: str,
    ) -> StorageVolume:
        storage_manager: StorageSessionManager = context['storage_manager']
        proxy_name, volume_name = storage_manager.split_host(id)
        try:
            proxy_info = storage_manager._proxies[proxy_name]
        except KeyError:
            raise ValueError(f"no such storage proxy: {proxy_name!r}")
        async with proxy_info.session.request(
            'GET', proxy_info.manager_api_url / 'volumes',
            raise_for_status=True,
            headers={AUTH_TOKEN_HDR: proxy_info.secret},
        ) as resp:
            reply = await resp.json()
            for volume_data in reply['volumes']:
                if volume_data['name'] == volume_name:
                    return cls.from_info(proxy_name, volume_data)
            else:
                raise ValueError(
                    f"no such volume in the storage proxy {proxy_name!r}: {volume_name!r}"
                )


class StorageVolumeList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList, )

    items = graphene.List(StorageVolume, required=True)



import inspect
import logging
from typing import (
    Any,
    Iterable,
    Tuple,
)
import re

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
import graphene
from graphql.execution.executors.asyncio import AsyncioExecutor
from graphql.error import GraphQLError, format_error
import trafaret as t

from ai.backend.common.logging import BraceStyleAdapter
from ai.backend.common import validators as tx

from .manager import GQLMutationUnfrozenRequiredMiddleware
from .exceptions import GraphQLError as BackendGQLError
from .auth import auth_required
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params
from ..manager.models.base import DataLoaderManager
from ..manager.models.gql import (
    Mutations, Queries,
    GQLMutationPrivilegeCheckMiddleware,
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.admin'))

_rx_qname = re.compile(r'{\s*(\w+)\b')


class GQLLoggingMiddleware:

    def resolve(self, next, root, info, **args):
        if len(info.path) == 1:
            log.info('ADMIN.GQL (ak:{}, {}:{}, op:{})',
                     info.context['access_key'],
                     info.operation.operation,
                     info.field_name,
                     info.operation.name)
        return next(root, info, **args)


@atomic
@auth_required
@check_api_params(
    t.Dict({
        t.Key('query'): t.String,
        t.Key('variables', default=None): t.Null | t.Mapping(t.String, t.Any),
        tx.AliasedKey(['operation_name', 'operationName'], default=None): t.Null | t.String,
    }))
async def handle_gql(request: web.Request, params: Any) -> web.Response:
    executor = request.app['admin.gql_executor']
    schema = request.app['admin.gql_schema']
    manager_status = await request.app['shared_config'].get_manager_status()
    known_slot_types = await request.app['shared_config'].get_resource_slots()
    context = {
        'local_config': request.app['local_config'],
        'shared_config': request.app['shared_config'],
        'etcd': request.app['shared_config'].etcd,
        'user': request['user'],
        'access_key': request['keypair']['access_key'],
        'dbpool': request.app['dbpool'],
        'redis_stat': request.app['redis_stat'],
        'manager_status': manager_status,
        'known_slot_types': known_slot_types,
        'background_task_manager': request.app['background_task_manager'],
        'storage_manager': request.app['storage_manager'],
        'registry': request.app['registry'],
    }
    dlmanager = DataLoaderManager(context)
    result = schema.execute(
        params['query'], executor,
        variable_values=params['variables'],
        operation_name=params['operation_name'],
        context_value={
            'dlmgr': dlmanager,
            **context,
        },
        middleware=[
            GQLLoggingMiddleware(),
            GQLMutationUnfrozenRequiredMiddleware(),
            GQLMutationPrivilegeCheckMiddleware(),
        ],
        return_promise=True)
    if inspect.isawaitable(result):
        result = await result
    if result.errors:
        errors = []
        for e in result.errors:
            if isinstance(e, GraphQLError):
                errmsg = format_error(e)
                errors.append(errmsg)
            else:
                errmsg = {'message': str(e)}
                errors.append(errmsg)
            log.error('ADMIN.GQL Exception: {}', errmsg)
        raise BackendGQLError(extra_data=errors)
    return web.json_response(result.data, status=200)


async def init(app: web.Application) -> None:
    app['admin.gql_executor'] = AsyncioExecutor()
    app['admin.gql_schema'] = graphene.Schema(
        query=Queries,
        mutation=Mutations,
        auto_camelcase=False)


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['api_versions'] = (2, 3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('POST', r'/graphql', handle_gql))
    return app, []


if __name__ == '__main__':
    # If executed as a main program, print all GraphQL schemas.
    # (graphene transforms our object model into a textual representation)
    # This is useful for writing documentation!
    schema = graphene.Schema(
        query=Queries,
        mutation=Mutations,
        auto_camelcase=False)
    print('======== GraphQL API Schema ========')
    print(str(schema))



from typing import (
    Union,
)

from lark import Lark
import lark
import sqlalchemy as sa
from sqlalchemy.sql import and_, or_, not_
from sqlalchemy.sql.elements import ClauseElement


FilterableSQLQuery = Union[sa.sql.Select, sa.sql.Update, sa.sql.Delete]


class QueryFilterParser():
    __JSON_GRAMMER = r"""
        ?start: value
        ?value: object
              | array
              | filter
              | string
              | pair
              | word
              | number
        array  : "[" [value ("," value)*] "]"
        object : "{" [value ("," value)*] "}"
        filter : [word|string] ":" object
        pair   : [word|string] ":" [word|string|number]
        word   : CNAME
        string : ESCAPED_STRING | /'[^']*'/
        number : SIGNED_NUMBER
        %import common.CNAME
        %import common.ESCAPED_STRING
        %import common.SIGNED_NUMBER
        %import common.WS
        %ignore WS
    """

    def __init__(self) -> None:
        self._FILTERS = ['and', 'or', 'not', 'gt', 'gte', 'lt', 'lte', 'eq', 'ne']
        self._JSON_PARSER = Lark(
            self.__JSON_GRAMMER,
            parser='lalr',
            maybe_placeholders=False,
        )

    def _parse_query(self, query: str) -> lark.tree.Tree:
        tree = self._JSON_PARSER.parse(query)
        return tree

    def _ast_to_where_clause(self, tree: lark.tree.Tree, table: sa.Table) -> ClauseElement:
        values_stack = []
        op_stack = []
        cond_stack = []
        for index, subtree in enumerate(tree.iter_subtrees_topdown()):
            data = subtree.data
            children = subtree.children
            if children and isinstance(children[0], lark.lexer.Token):
                token = str(children[0])
                if token in self._FILTERS:
                    op_stack.append(token)
                else:
                    if data == "word":
                        if token not in table.c:
                            raise ValueError("Unknown column name in the table", token)
                        values_stack.append(table.c[token])
                    elif data == "string":
                        values_stack.append(token[1:-1])
                    else:
                        values_stack.append(token)
        while op_stack:
            op = op_stack.pop()
            if op == "not":
                cond = not_(values_stack.pop())
                cond_stack.append(cond)
                continue
            if op in ["and", "or"]:
                value1 = cond_stack.pop()
                value2 = cond_stack.pop()
            else:
                value2 = values_stack.pop()
                value1 = values_stack.pop()
            if op == "gt":
                cond = value1 > value2
            elif op == "gte":
                cond = value1 >= value2
            elif op == "lt":
                cond = value1 < value2
            elif op == "lte":
                cond = value1 <= value2
            elif op == "eq":
                cond = value1 == value2
            elif op == "ne":
                cond = value1 != value2
            elif op == "and":
                cond = and_(value1, value2)
            elif op == "or":
                cond = or_(value1, value2)
            cond_stack.append(cond)
        return cond_stack.pop()

    def append_filter(
        self,
        sa_query: FilterableSQLQuery,
        filter_expr: str,
    ) -> FilterableSQLQuery:
        """
        Parse the given filter expression and build the where clause based on the first target table from
        the given SQLAlchemy query object.
        """
        if isinstance(sa_query, sa.sql.Select):
            table = sa_query.selectable.locate_all_froms()[0]
        elif isinstance(sa_query, sa.sql.Delete):
            table = sa_query.table
        elif isinstance(sa_query, sa.sql.Update):
            table = sa_query.table
        else:
            raise ValueError('Unsupported SQLAlchemy query object type')
        ast = self._parse_query(filter_expr)
        if ast.data != "filter":
            raise ValueError("The form of the query_arg is not a filter")
        where_clause = self._ast_to_where_clause(ast, table)
        return sa_query.where(where_clause)



"""
This module defines a series of Backend.AI-specific errors based on HTTP Error
classes from aiohttp.
Raising a BackendError is automatically mapped to a corresponding HTTP error
response with RFC7807-style JSON-encoded description in its response body.

In the client side, you should use "type" field in the body to distinguish
canonical error types beacuse "title" field may change due to localization and
future UX improvements.
"""

from __future__ import annotations

import json
from typing import (
    Any,
    Dict,
    Optional,
    Mapping,
    Union,
    cast,
)

from aiohttp import web

from ..manager.exceptions import AgentError

from ai.backend.common.plugin.hook import HookResult


class BackendError(web.HTTPError):
    """
    An RFC-7807 error class as a drop-in replacement of the original
    aiohttp.web.HTTPError subclasses.
    """

    error_type: str  = 'https://api.backend.ai/probs/general-error'
    error_title: str = 'General Backend API Error.'

    content_type: str
    extra_msg: Optional[str]

    def __init__(self, extra_msg: str = None, extra_data: Any = None, **kwargs):
        super().__init__(**kwargs)
        self.args = (self.status_code, self.reason, self.error_type)
        self.empty_body = False
        self.content_type = 'application/problem+json'
        self.extra_msg = extra_msg
        self.extra_data = extra_data
        body = {
            'type': self.error_type,
            'title': self.error_title,
        }
        if extra_msg is not None:
            body['msg'] = extra_msg
        if extra_data is not None:
            body['data'] = extra_data
        self.body = json.dumps(body).encode()

    def __str__(self):
        lines = []
        if self.extra_msg:
            lines.append(f'{self.error_title} ({self.extra_msg})')
        else:
            lines.append(self.error_title)
        if self.extra_data:
            lines.append(' -> extra_data: ' + repr(self.extra_data))
        return '\n'.join(lines)

    def __repr__(self):
        lines = []
        if self.extra_msg:
            lines.append(f'<{type(self).__name__}({self.status}): '
                         f'{self.error_title} ({self.extra_msg})>')
        else:
            lines.append(f'<{type(self).__name__}({self.status}): '
                         f'{self.error_title}>')
        if self.extra_data:
            lines.append(' -> extra_data: ' + repr(self.extra_data))
        return '\n'.join(lines)

    def __reduce__(self):
        return (
            type(self),
            (),  # empty the constructor args to make unpickler to use
                 # only the exact current state in __dict__
            self.__dict__,
        )


class GenericNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/generic-not-found'
    error_title = 'Unknown URL path.'


class GenericBadRequest(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/generic-bad-request'
    error_title = 'Bad request.'


class RejectedByHook(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/rejected-by-hook'
    error_title = 'Operation rejected by a hook plugin.'

    @classmethod
    def from_hook_result(cls, result: HookResult) -> RejectedByHook:
        return cls(
            extra_msg=result.reason,
            extra_data={
                'plugins': result.src_plugin,
            },
        )


class InvalidCredentials(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/invalid-credentials'
    error_title = 'Invalid credentials for authentication.'


class GenericForbidden(BackendError, web.HTTPForbidden):
    error_type  = 'https://api.backend.ai/probs/generic-forbidden'
    error_title = 'Forbidden operation.'


class InsufficientPrivilege(BackendError, web.HTTPForbidden):
    error_type  = 'https://api.backend.ai/probs/insufficient-privilege'
    error_title = 'Insufficient privilege.'


class MethodNotAllowed(BackendError, web.HTTPMethodNotAllowed):
    error_type  = 'https://api.backend.ai/probs/method-not-allowed'
    error_title = 'HTTP Method Not Allowed.'


class InternalServerError(BackendError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/internal-server-error'
    error_title = 'Internal server error.'


class ServerMisconfiguredError(BackendError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/server-misconfigured'
    error_title = 'Service misconfigured.'


class ServiceUnavailable(BackendError, web.HTTPServiceUnavailable):
    error_type  = 'https://api.backend.ai/probs/service-unavailable'
    error_title = 'Serivce unavailable.'


class QueryNotImplemented(BackendError, web.HTTPServiceUnavailable):
    error_type  = 'https://api.backend.ai/probs/not-implemented'
    error_title = 'This API query is not implemented.'


class InvalidAuthParameters(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/invalid-auth-params'
    error_title = 'Missing or invalid authorization parameters.'


class AuthorizationFailed(BackendError, web.HTTPUnauthorized):
    error_type  = 'https://api.backend.ai/probs/auth-failed'
    error_title = 'Credential/signature mismatch.'


class InvalidAPIParameters(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/invalid-api-params'
    error_title = 'Missing or invalid API parameters.'


class GraphQLError(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/graphql-error'
    error_title = 'GraphQL-generated error.'


class InstanceNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/instance-not-found'
    error_title = 'No such instance.'


class ImageNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/image-not-found'
    error_title = 'No such environment image.'


class DomainNotFound(web.HTTPNotFound, BackendError):
    error_type = 'https://api.backend.ai/probs/domain-not-found'
    error_title = 'No such domain.'


class GroupNotFound(web.HTTPNotFound, BackendError):
    error_type  = 'https://api.backend.ai/probs/group-not-found'
    error_title = 'No such user group.'


class ScalingGroupNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/scaling-group-not-found'
    error_title = 'No such scaling group.'


class SessionNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/session-not-found'
    error_title = 'No such session.'


class TooManySessionsMatched(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/too-many-sessions-matched'
    error_title = 'Too many sessions matched.'

    def __init__(self, extra_msg: str = None, extra_data: Dict[str, Any] = None, **kwargs):
        if (
            extra_data is not None and
            (matches := extra_data.get('matches', None)) is not None
        ):
            serializable_matches = [{
                'id': str(item['session_id']),
                'name': item['session_name'],
                'status': item['status'].name,
                'created_at': item['created_at'].isoformat(),
            } for item in matches]
            extra_data['matches'] = serializable_matches
        super().__init__(extra_msg, extra_data, **kwargs)


class TaskTemplateNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/task-template-not-found'
    error_title = 'No such task template.'


class TooManyKernelsFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/too-many-kernels'
    error_title = 'There are two or more matching kernels.'


class AppNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/app-not-found'
    error_title = 'No such app service provided by the session.'


class SessionAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/session-already-exists'
    error_title = 'The session already exists but you requested not to reuse existing one.'


class VFolderCreationFailed(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/vfolder-creation-failed'
    error_title = 'Virtual folder creation has failed.'


class VFolderNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/vfolder-not-found'
    error_title = 'No such virtual folder.'


class VFolderAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/vfolder-already-exists'
    error_title = 'The virtual folder already exists with the same name.'


class VFolderOperationFailed(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/vfolder-operation-failed'
    error_title = 'Virtual folder operation has failed.'


class DotfileCreationFailed(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/generic-bad-request'
    error_title = 'Dotfile creation has failed.'


class DotfileAlreadyExists(BackendError, web.HTTPBadRequest):
    error_type  = 'https://api.backend.ai/probs/generic-bad-request'
    error_title = 'Dotfile already exists.'


class DotfileNotFound(BackendError, web.HTTPNotFound):
    error_type  = 'https://api.backend.ai/probs/generic-not-found'
    error_title = 'Requested Dotfile not found.'


class QuotaExceeded(BackendError, web.HTTPPreconditionFailed):
    error_type  = 'https://api.backend.ai/probs/quota-exceeded'
    error_title = 'You have reached your resource limit.'


class RateLimitExceeded(BackendError, web.HTTPTooManyRequests):
    error_type  = 'https://api.backend.ai/probs/rate-limit-exceeded'
    error_title = 'You have reached your API query rate limit.'


class InstanceNotAvailable(BackendError, web.HTTPServiceUnavailable):
    error_type  = 'https://api.backend.ai/probs/instance-not-available'
    error_title = 'There is no available instance.'


class ServerFrozen(BackendError, web.HTTPServiceUnavailable):
    error_type  = 'https://api.backend.ai/probs/server-frozen'
    error_title = 'The server is frozen due to maintenance. Please try again later.'


class StorageProxyError(BackendError, web.HTTPError):
    error_type  = 'https://api.backend.ai/probs/storage-proxy-error'
    error_title = 'The storage proxy returned an error.'

    def __init__(self, status: int, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Currently there is no good public way to override the status code
        # after initialization of aiohttp.web.StreamResponse objects. :(
        self.status_code = status  # HTTPException uses self.status_code
        self._status = status      # StreamResponse uses self._status
        self.args = (status, self.args[1], self.args[2])

    @property
    def status(self) -> int:
        # override the status property again to refer the subclass' attribute.
        return self.status_code


class BackendAgentError(BackendError):
    """
    An RFC-7807 error class that wraps agent-side errors.
    """

    _short_type_map = {
        'TIMEOUT': 'https://api.backend.ai/probs/agent-timeout',
        'INVALID_INPUT': 'https://api.backend.ai/probs/agent-invalid-input',
        'FAILURE': 'https://api.backend.ai/probs/agent-failure',
    }

    def __init__(self, agent_error_type: str,
                 exc_info: Union[str, AgentError, Exception, Mapping[str, Optional[str]], None] = None):
        super().__init__()
        agent_details: Mapping[str, Optional[str]]
        if not agent_error_type.startswith('https://'):
            agent_error_type = self._short_type_map[agent_error_type.upper()]
        self.args = (
            self.status_code,
            self.reason,
            self.error_type,
            agent_error_type,
        )
        if isinstance(exc_info, str):
            agent_details = {
                'type': agent_error_type,
                'title': exc_info,
            }
        elif isinstance(exc_info, AgentError):
            e = cast(AgentError, exc_info)
            agent_details = {
                'type': agent_error_type,
                'title': 'Agent-side exception occurred.',
                'exception': e.exc_repr,
            }
        elif isinstance(exc_info, Exception):
            agent_details = {
                'type': agent_error_type,
                'title': 'Unexpected exception ocurred.',
                'exception': repr(exc_info),
            }
        elif isinstance(exc_info, Mapping):
            agent_details = exc_info
        else:
            agent_details = {
                'type': agent_error_type,
                'title': None if exc_info is None else str(exc_info),
            }
        self.agent_details = agent_details
        self.agent_error_type = agent_error_type
        self.agent_error_title = agent_details['title']
        self.agent_exception = agent_details.get('exception', '')
        self.body = json.dumps({
            'type': self.error_type,
            'title': self.error_title,
            'agent-details': agent_details,
        }).encode()

    def __str__(self):
        if self.agent_exception:
            return f'{self.agent_error_title} ({self.agent_exception})'
        return f'{self.agent_error_title}'

    def __repr__(self):
        if self.agent_exception:
            return f'<{type(self).__name__}: {self.agent_error_title} ({self.agent_exception})>'
        return f'<{type(self).__name__}: {self.agent_error_title}>'

    def __reduce__(self):
        return (type(self), (self.agent_error_type, self.agent_details))


class KernelCreationFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/kernel-creation-failed'
    error_title = 'Kernel creation has failed.'


class KernelDestructionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/kernel-destruction-failed'
    error_title = 'Kernel destruction has failed.'


class KernelRestartFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/kernel-restart-failed'
    error_title = 'Kernel restart has failed.'


class KernelExecutionFailed(BackendAgentError, web.HTTPInternalServerError):
    error_type  = 'https://api.backend.ai/probs/kernel-execution-failed'
    error_title = 'Executing user code in the kernel has failed.'



"""convert_cpu_gpu_slots_to_float

Revision ID: d52bf5ec9ef3
Revises: 4545f5c948b3
Create Date: 2017-11-09 14:30:20.737908

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd52bf5ec9ef3'
down_revision = '4545f5c948b3'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column('agents', 'mem_slots',
                    existing_type=sa.Integer(),
                    type_=sa.BigInteger())
    op.alter_column('agents', 'cpu_slots',
                    existing_type=sa.Integer(),
                    type_=sa.Float())
    op.alter_column('agents', 'gpu_slots',
                    existing_type=sa.Integer(),
                    type_=sa.Float())
    op.alter_column('agents', 'used_mem_slots',
                    existing_type=sa.Integer(),
                    type_=sa.BigInteger())
    op.alter_column('agents', 'used_cpu_slots',
                    existing_type=sa.Integer(),
                    type_=sa.Float())
    op.alter_column('agents', 'used_gpu_slots',
                    existing_type=sa.Integer(),
                    type_=sa.Float())
    op.alter_column('kernels', 'mem_slot',
                    existing_type=sa.Integer(),
                    type_=sa.BigInteger())
    op.alter_column('kernels', 'cpu_slot',
                    existing_type=sa.Integer(),
                    type_=sa.Float())
    op.alter_column('kernels', 'gpu_slot',
                    existing_type=sa.Integer(),
                    type_=sa.Float())


def downgrade():
    op.alter_column('agents', 'mem_slots',
                    existing_type=sa.BigInteger(),
                    type_=sa.Integer())
    op.alter_column('agents', 'cpu_slots',
                    existing_type=sa.Float(),
                    type_=sa.Integer())
    op.alter_column('agents', 'gpu_slots',
                    existing_type=sa.Float(),
                    type_=sa.Integer())
    op.alter_column('agents', 'used_mem_slots',
                    existing_type=sa.BigInteger(),
                    type_=sa.Integer())
    op.alter_column('agents', 'used_cpu_slots',
                    existing_type=sa.Float(),
                    type_=sa.Integer())
    op.alter_column('agents', 'used_gpu_slots',
                    existing_type=sa.Float(),
                    type_=sa.Integer())
    op.alter_column('kernels', 'mem_slot',
                    existing_type=sa.BigInteger(),
                    type_=sa.Integer())
    op.alter_column('kernels', 'cpu_slot',
                    existing_type=sa.Float(),
                    type_=sa.Integer())
    op.alter_column('kernels', 'gpu_slot',
                    existing_type=sa.Float(),
                    type_=sa.Integer())



import logging
import re
from typing import Any, Tuple

from aiohttp import web
import aiohttp_cors
import trafaret as t

from ai.backend.common import msgpack
from ai.backend.common.logging import BraceStyleAdapter

from .auth import auth_required
from .exceptions import (
    InvalidAPIParameters, DotfileCreationFailed,
    DotfileNotFound, DotfileAlreadyExists
)
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, Iterable, WebMiddleware
from .utils import check_api_params, get_access_key_scopes

from ..manager.models import (
    keypairs,
    vfolders,
    query_accessible_vfolders,
    query_bootstrap_script,
    query_owned_dotfiles,
    verify_dotfile_name,
    MAXIMUM_DOTFILE_SIZE,
)

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.dotfile'))


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(t.Dict(
    {
        t.Key('data'): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
        t.Key('path'): t.String,
        t.Key('permission'): t.Regexp(r'^[0-7]{3}$', re.ASCII),
        t.Key('owner_access_key', default=None): t.Null | t.String,
    }
))
async def create(request: web.Request, params: Any) -> web.Response:
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info('CREATE (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')
    dbpool = request.app['dbpool']
    user_uuid = request['user']['uuid']
    async with dbpool.acquire() as conn, conn.begin():
        path: str = params['path']
        dotfiles, leftover_space = await query_owned_dotfiles(conn, owner_access_key)
        if leftover_space == 0:
            raise DotfileCreationFailed('No leftover space for dotfile storage')
        if len(dotfiles) == 100:
            raise DotfileCreationFailed('Dotfile creation limit reached')
        if not verify_dotfile_name(path):
            raise InvalidAPIParameters('dotfile path is reserved for internal operations.')
        duplicate_vfolder = \
            await query_accessible_vfolders(conn, user_uuid, extra_vf_conds=(vfolders.c.name == path))
        if len(duplicate_vfolder) > 0:
            raise InvalidAPIParameters('dotfile path conflicts with your dot-prefixed vFolder')
        duplicate = [x for x in dotfiles if x['path'] == path]
        if len(duplicate) > 0:
            raise DotfileAlreadyExists
        new_dotfiles = list(dotfiles)
        new_dotfiles.append({'path': path, 'perm': params['permission'], 'data': params['data']})
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed('No leftover space for dotfile storage')

        query = (keypairs.update()
                         .values(dotfiles=dotfile_packed)
                         .where(keypairs.c.access_key == owner_access_key))
        await conn.execute(query)
    return web.json_response({})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(t.Dict({
    t.Key('path', default=None): t.Null | t.String,
    t.Key('owner_access_key', default=None): t.Null | t.String,
}))
async def list_or_get(request: web.Request, params: Any) -> web.Response:
    resp = []
    dbpool = request.app['dbpool']
    access_key = request['keypair']['access_key']

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info('LIST_OR_GET (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')
    async with dbpool.acquire() as conn:
        if params['path']:
            dotfiles, _ = await query_owned_dotfiles(conn, owner_access_key)
            for dotfile in dotfiles:
                if dotfile['path'] == params['path']:
                    return web.json_response(dotfile)
            raise DotfileNotFound
        else:
            dotfiles, _ = await query_owned_dotfiles(conn, access_key)
            for entry in dotfiles:
                resp.append({
                    'path': entry['path'],
                    'permission': entry['perm'],
                    'data': entry['data']
                })
            return web.json_response(resp)


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(t.Dict(
    {
        t.Key('data'): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
        t.Key('path'): t.String,
        t.Key('permission'): t.Regexp(r'^[0-7]{3}$', re.ASCII),
        t.Key('owner_access_key', default=None): t.Null | t.String,
    }
))
async def update(request: web.Request, params: Any) -> web.Response:
    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info('CREATE (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')
    dbpool = request.app['dbpool']

    async with dbpool.acquire() as conn, conn.begin():
        path: str = params['path']
        dotfiles, _ = await query_owned_dotfiles(conn, owner_access_key)
        new_dotfiles = [x for x in dotfiles if x['path'] != path]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound

        new_dotfiles.append({'path': path, 'perm': params['permission'], 'data': params['data']})
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed('No leftover space for dotfile storage')

        query = (keypairs.update()
                         .values(dotfiles=dotfile_packed)
                         .where(keypairs.c.access_key == owner_access_key))
        await conn.execute(query)
    return web.json_response({})


@auth_required
@server_status_required(READ_ALLOWED)
@check_api_params(
    t.Dict({
        t.Key('path'): t.String,
        t.Key('owner_access_key', default=None): t.Null | t.String,
    })
)
async def delete(request: web.Request, params: Any) -> web.Response:
    dbpool = request.app['dbpool']
    path = params['path']

    requester_access_key, owner_access_key = await get_access_key_scopes(request, params)
    log.info('DELETE (ak:{0}/{1})',
             requester_access_key, owner_access_key if owner_access_key != requester_access_key else '*')

    async with dbpool.acquire() as conn, conn.begin():
        dotfiles, _ = await query_owned_dotfiles(conn, owner_access_key)
        new_dotfiles = [x for x in dotfiles if x['path'] != path]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound
        dotfile_packed = msgpack.packb(new_dotfiles)
        query = (keypairs.update()
                         .values(dotfiles=dotfile_packed)
                         .where(keypairs.c.access_key == owner_access_key))
        await conn.execute(query)
        return web.json_response({'success': True})


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(t.Dict(
    {
        t.Key('script'): t.String(allow_blank=True, max_length=MAXIMUM_DOTFILE_SIZE),
    }
))
async def update_bootstrap_script(request: web.Request, params: Any) -> web.Response:
    dbpool = request.app['dbpool']
    access_key = request['keypair']['access_key']
    log.info('UPDATE_BOOTSTRAP_SCRIPT (ak:{0})', access_key)
    async with dbpool.acquire() as conn, conn.begin():
        script = params.get('script', '').strip()
        if len(script) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed('Maximum bootstrap script length reached')
        query = (keypairs.update()
                         .values(bootstrap_script=script)
                         .where(keypairs.c.access_key == access_key))
        await conn.execute(query)
    return web.json_response({})


@auth_required
@server_status_required(READ_ALLOWED)
async def get_bootstrap_script(request: web.Request) -> web.Response:
    dbpool = request.app['dbpool']
    access_key = request['keypair']['access_key']
    log.info('GET_BOOTSTRAP_SCRIPT (ak:{0})', access_key)
    async with dbpool.acquire() as conn:
        script, _ = await query_bootstrap_script(conn, access_key)
        return web.json_response(script)


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app['api_versions'] = (4, 5)
    app['prefix'] = 'user-config'
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('POST', '/dotfiles', create))
    cors.add(app.router.add_route('GET', '/dotfiles', list_or_get))
    cors.add(app.router.add_route('PATCH', '/dotfiles', update))
    cors.add(app.router.add_route('DELETE', '/dotfiles', delete))
    cors.add(app.router.add_route('POST', '/bootstrap-script', update_bootstrap_script))
    cors.add(app.router.add_route('GET', '/bootstrap-script', get_bootstrap_script))

    return app, []



from __future__ import annotations

import logging
from typing import (
    Any,
    AsyncGenerator,
    Iterable,
    Mapping,
    TYPE_CHECKING,
    Tuple,
)

from aiohttp import web
import aiohttp_cors
from aiojobs.aiohttp import atomic
import trafaret as t

from ai.backend.common.docker import get_known_registries
from ai.backend.common.logging import BraceStyleAdapter
from .auth import superadmin_required
from .exceptions import InvalidAPIParameters
from .utils import check_api_params
from .types import CORSOptions, WebMiddleware
if TYPE_CHECKING:
    from .config import SharedConfig

log = BraceStyleAdapter(logging.getLogger('ai.backend.gateway.etcd'))


@atomic
async def get_resource_slots(request: web.Request) -> web.Response:
    log.info('ETCD.GET_RESOURCE_SLOTS ()')
    shared_config: SharedConfig = request.app['shared_config']
    known_slots = await shared_config.get_resource_slots()
    return web.json_response(known_slots, status=200)


@atomic
async def get_vfolder_types(request: web.Request) -> web.Response:
    log.info('ETCD.GET_VFOLDER_TYPES ()')
    shared_config: SharedConfig = request.app['shared_config']
    vfolder_types = await shared_config.get_vfolder_types()
    return web.json_response(vfolder_types, status=200)


@atomic
@superadmin_required
async def get_docker_registries(request: web.Request) -> web.Response:
    """
    Returns the list of all registered docker registries.
    """
    log.info('ETCD.GET_DOCKER_REGISTRIES ()')
    etcd = request.app['shared_config'].etcd
    _registries = await get_known_registries(etcd)
    # ``yarl.URL`` is not JSON-serializable, so we need to represent it as string.
    known_registries: Mapping[str, str] = {k: v.human_repr() for k, v in _registries.items()}
    return web.json_response(known_registries, status=200)


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('key'): t.String,
        t.Key('prefix', default=False): t.Bool,
    }))
async def get_config(request: web.Request, params: Any) -> web.Response:
    shared_config: SharedConfig = request.app['shared_config']
    etcd = shared_config.etcd
    log.info('ETCD.GET_CONFIG (ak:{}, key:{}, prefix:{})',
             request['keypair']['access_key'], params['key'], params['prefix'])
    if params['prefix']:
        # Flatten the returned ChainMap object for JSON serialization
        value = dict(await etcd.get_prefix_dict(params['key']))
    else:
        value = await etcd.get(params['key'])
    return web.json_response({'result': value})


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('key'): t.String,
        t.Key('value'): (t.String(allow_blank=True) |
                         t.Mapping(t.String(allow_blank=True), t.Any)),
    }))
async def set_config(request: web.Request, params: Any) -> web.Response:
    shared_config: SharedConfig = request.app['shared_config']
    etcd = shared_config.etcd
    log.info('ETCD.SET_CONFIG (ak:{}, key:{}, val:{})',
             request['keypair']['access_key'], params['key'], params['value'])
    if isinstance(params['value'], Mapping):
        updates = {}

        def flatten(prefix, o):
            for k, v in o.items():
                inner_prefix = prefix if k == '' else f'{prefix}/{k}'
                if isinstance(v, Mapping):
                    flatten(inner_prefix, v)
                else:
                    updates[inner_prefix] = v

        flatten(params['key'], params['value'])
        # TODO: chunk support if there are too many keys
        if len(updates) > 16:
            raise InvalidAPIParameters(
                'Too large update! Split into smaller key-value pair sets.')
        await etcd.put_dict(updates)
    else:
        await etcd.put(params['key'], params['value'])
    return web.json_response({'result': 'ok'})


@atomic
@superadmin_required
@check_api_params(
    t.Dict({
        t.Key('key'): t.String,
        t.Key('prefix', default=False): t.Bool,
    }))
async def delete_config(request: web.Request, params: Any) -> web.Response:
    shared_config: SharedConfig = request.app['shared_config']
    etcd = shared_config.etcd
    log.info('ETCD.DELETE_CONFIG (ak:{}, key:{}, prefix:{})',
             request['keypair']['access_key'], params['key'], params['prefix'])
    if params['prefix']:
        await etcd.delete_prefix(params['key'])
    else:
        await etcd.delete(params['key'])
    return web.json_response({'result': 'ok'})


async def app_ctx(app: web.Application) -> AsyncGenerator[None, None]:
    if app['pidx'] == 0:
        await app['shared_config'].register_myself()
    yield
    if app['pidx'] == 0:
        await app['shared_config'].deregister_myself()


def create_app(default_cors_options: CORSOptions) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.cleanup_ctx.append(app_ctx)
    app['prefix'] = 'config'
    app['api_versions'] = (3, 4)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route('GET',  r'/resource-slots', get_resource_slots))
    cors.add(app.router.add_route('GET',  r'/vfolder-types', get_vfolder_types))
    cors.add(app.router.add_route('GET',  r'/docker-registries', get_docker_registries))
    cors.add(app.router.add_route('POST', r'/get', get_config))
    cors.add(app.router.add_route('POST', r'/set', set_config))
    cors.add(app.router.add_route('POST', r'/delete', delete_config))
    return app, []
